"""
    Some basic utility methods.
"""
from collections import OrderedDict

from .errors import DuplicateDefinition


class IndexDictionary:
    """ A basic indexing object that assigns consecutive indexes to the indexed objects. """
    def __init__(self, elements=None):
        self.data = OrderedDict()
        self.objects = []
        _ = [self.add(element) for element in (elements or [])]

    def get_index(self, key):
        return self.data[key]

    def get_object(self, index):
        return self.objects[index]

    def add(self, obj):
        if obj in self.data:
            raise DuplicateDefinition(obj, self.data[obj])
        self.data[obj] = len(self.data)
        self.objects.append(obj)

    def dump(self):
        return [str(o) for o in self.data.keys()]

    def __str__(self):
        return ','.join('{}: {}'.format(idx, o) for o, idx in self.data.items())

    __repr__ = __str__

    def __iter__(self):
        return self.data.__iter__()

    def enumerate(self):
        """ Iterate over all (ordered) pairs of the form idx, o, where idx is the index of object 'o' """
        return ((idx, o) for o, idx in self.data.items())

    def __contains__(self, k):
        return k in self.data

    def __len__(self):
        return len(self.data)




from .formulas import Formula
from .terms import Term


def symref(sym):
    """ Create a reference of a logical expression, formula or term """
    if isinstance(sym, Term):
        return TermReference(sym)
    if isinstance(sym, Formula):
        # Formulas are hashable and indexable by themselves, but we wrap them in a Ref object for uniformity reasons
        return TermReference(sym)
    raise RuntimeError("Unexpected symbol type \"{}\"".format(type(sym)))


class TermReference:
    """ A simple wrapper to provide a purely syntactic __eq__ operator for terms.
     To be used whenever equality and hashing is required, e.g. in dictionaries, etc.,
     since the __eq__ operator in the Term hierarchy is used for other purposes,
     namely, to construct equality atoms in a user-readable manner. """

    def __init__(self, expression):
        self.expr = expression

    def __hash__(self):
        return self.expr.hash()

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.expr.is_syntactically_equal(other.expr)

    def __str__(self):
        return "symref[{}]".format(self.expr)

    __repr__ = __str__



from ...syntax import CompoundFormula, Atom, Connective
from . import errors as err


class Sensor:
    """ A (possibly lifted) sensing action """

    def __init__(self, language, name, parameters, condition, obs):
        self.name = name
        self.language = language
        self.parameters = parameters
        self.condition = condition
        self.obs = obs
        self._check_well_formed()

    def _check_well_formed(self):
        if isinstance(self.obs, Atom):
            return True
        if isinstance(self.obs, CompoundFormula):
            if self.obs.connective == Connective.Not:
                if isinstance(self.obs.subformulas[0], Atom):
                    return True
        # We do not support anything other than literals
        raise err.ObservationExpressivenessMismatch(self)

    def ident(self):
        params = ', '.join([str(o) for o in self.parameters])
        return '{}({})'.format(self.name, params)

    def dump(self):
        return dict(name=self.name,
                    params=[par.dump() for par in self.parameters],
                    condition=self.condition.dump(),
                    obs=[eff.dump() for eff in self.effects.dump()])

    def __str__(self):
        tokens = ['action {}:'.format(self.name),
                  'C=({})'.format(self.condition),
                  'L=({})'.format(str(self.obs))]
        return '\n'.join(tokens)




from .problem import Problem, create_fstrips_problem
from .action import Action
from .derived import Derived
from .fstrips import SingleEffect, AddEffect, DelEffect, FunctionalEffect, IncreaseEffect, \
    LiteralEffect, UniversalEffect, ChoiceEffect, VectorisedEffect, LinearEffect, BlackBoxEffect,\
    language, OptimizationMetric, OptimizationType




from collections import OrderedDict
from ..problem import Problem
from .sensor import Sensor
from . import errors as err


class ContingentProblem(Problem):
    """ A Functional STRIPS Contingent problem """

    def __init__(self):
        super().__init__()
        self.sensors = OrderedDict()

    def sensor(self, name, parameters, condition, obs):
        if name in self.sensors:
            raise err.DuplicateSensorDefinition(name, self.sensors[name])
        self.sensors[name] = Sensor(self.language, name, parameters, condition, obs)
        return self.sensors[name]

    def has_sensor(self, name):
        return name in self.sensors

    def get_sensor(self, name):
        if not self.has_sensor(name):
            raise err.UndefinedSensor(name)
        return self.sensors[name]

    def get_symbols(self, pv, ev, cv):
        super().get_symbols(pv, ev, cv)
        for _, sensor in self.sensors.items():
            pv.visit(sensor.condition)
            pv.visit(sensor.obs)

    def __str__(self):
        return 'FSTRIPS Contingent Problem "{}", domain "{}"'.format(self.name, self.domain_name)

    __repr__ = __str__




from .concepts import Concept, PrimitiveConcept, UniversalConcept, NotConcept, ExistsConcept, ForallConcept, \
    EqualConcept, AndConcept, OrConcept, EmptyConcept, NominalConcept, NullaryAtom, GoalConcept, GoalNullaryAtom
from .concepts import Role, PrimitiveRole, InverseRole, StarRole, RestrictRole, CompositionRole, GoalRole
from .features import MinDistanceFeature, ConceptCardinalityFeature, EmpiricalBinaryConcept, FeatureValueChange,\
    NullaryAtomFeature
from .factory import SyntacticFactory, compute_dl_vocabulary
from .errors import ArityDLMismatch




from .function import Function
from .predicate import Predicate
from .sorts import Sort, Interval, inclusion_closure
from .terms import Term, Constant, Variable, CompoundTerm, IfThenElse, ite, AggregateCompoundTerm
from .util import termlists_are_equal, termlist_hash
from .formulas import land, lor, neg, implies, forall, exists, equiv, Connective, Atom, Formula,\
    CompoundFormula, QuantifiedFormula, Tautology, Contradiction, top, bot, Quantifier, VariableBinding
from .builtins import BuiltinFunctionSymbol, BuiltinPredicateSymbol
from .symrefs import symref
from .transform.subst import create_substitution, term_substitution, TermSubstitution



"""
    RDDL model acquisition module
"""
from enum import Enum

from pyrddl.parser import RDDLParser

from .common import load_tpl
from ..fol import FirstOrderLanguage
from ..syntax import implies, land, lor, neg, Connective, Quantifier, CompoundTerm, Interval, Atom, IfThenElse, \
    Contradiction, Tautology, CompoundFormula, forall, ite, AggregateCompoundTerm, QuantifiedFormula, Term, Function, \
    Variable, Predicate, Constant, Formula, builtins
from ..syntax import arithmetic as tm
from ..syntax.temporal import ltl as tt
from ..syntax.builtins import create_atom, BuiltinPredicateSymbol as BPS, BuiltinFunctionSymbol as BFS
from ..model import Model
from ..evaluators.simple import evaluate
from ..errors import LanguageError
from ..theories import Theory, language


class TranslationError(Exception):
    pass


logic_rddl_to_tarski = {
    '=>': implies,
    '^': land,
    '|': lor,
    '~': neg}

relational_rddl_to_tarski = {
    '==': BPS.EQ,
    '<=': BPS.LE,
    '>=': BPS.GE
}

arithmetic_rddl_to_tarski = {
    '-': lambda lang, x, y: lang.dispatch_operator(BFS.SUB, Term, Term, x, y),
    '+': lambda lang, x, y: lang.dispatch_operator(BFS.ADD, Term, Term, x, y),
    '*': lambda lang, x, y: lang.dispatch_operator(BFS.MUL, Term, Term, x, y),
    '/': lambda lang, x, y: lang.dispatch_operator(BFS.DIV, Term, Term, x, y)
}

func_rddl_to_tarski = {
    'abs': BFS.ABS,
    'Normal': BFS.NORMAL,
    'sqrt': BFS.SQRT,
    'pow': BFS.POW,
    'exp': BFS.EXP,
    'max': BFS.MAX,
    'Gamma': BFS.GAMMA
}


def translate_expression(lang, rddl_expr):
    try:
        expr_type, expr_sym = rddl_expr.etype
    except AttributeError:
        expr_type, expr_sym = rddl_expr

    if expr_type == 'number':
        if 'float' in expr_sym:
            return Constant(rddl_expr.args, lang.Real)
        elif 'int' in expr_sym:
            return Constant(rddl_expr.args, lang.Integer)
    elif expr_type == 'pvar_expr':
        # MRJ: this is a next state fluent
        # prima_fluent = False
        pvar_name = expr_sym[0]
        # if "'" in pvar_name:
        #     prima_fluent = True
        tsym = lang.get(pvar_name.replace("'", ""))
        args = expr_sym[1]
        targs = []
        if args is not None:
            for k, arg in enumerate(args):
                if arg[0] == '?':
                    if isinstance(tsym, Function):
                        targs += [Variable(arg, tsym.domain[k])]
                    elif isinstance(tsym, Predicate):
                        targs += [Variable(arg, tsym.sort[k])]
                    else:
                        assert False
                elif isinstance(arg, float):
                    targs += [Constant(arg, lang.Real)]
                elif isinstance(arg, int):
                    targs += [Constant(arg, lang.Integer)]
                else:
                    # print(arg)
                    targs += [lang.get(arg)]  # named constant?
        return tsym(*targs)

    elif expr_type == 'pvar':
        # MRJ: this is a next state fluent
        prima_fluent = False
        if "'" in rddl_expr.args[0]:
            prima_fluent = True
        tsym = lang.get(rddl_expr.args[0].replace("'", ""))
        if len(rddl_expr.args) == 2 and rddl_expr.args[1] is None:
            if prima_fluent:
                if not isinstance(tsym, Predicate):
                    raise LanguageError("Temporal operator X only defined for predicates!")
                return tt.X(tsym())
            return tsym()  # 0-ary
        targs = []
        for k, arg in enumerate(rddl_expr.args[1]):
            if arg[0] == '?':
                if isinstance(tsym, Function):
                    targs += [Variable(arg, tsym.domain[k])]
                elif isinstance(tsym, Predicate):
                    targs += [Variable(arg, tsym.sort[k])]
                else:
                    assert False
            elif isinstance(arg, float):
                targs += [Constant(arg, lang.Real)]
            elif isinstance(arg, int):
                targs += [Constant(arg, lang.Integer)]
            else:
                # print(arg)
                targs += [lang.get(arg)]  # named constant?
        if prima_fluent:
            if not isinstance(tsym, Predicate):
                raise LanguageError("Temporal operator X only defined for predicates!")
            return tt.X(tsym(*targs))

        return tsym(*targs)
    elif expr_type == 'typed_var':
        var_name, var_type = expr_sym
        return Variable(var_name, lang.get(var_type))
    elif expr_type == 'control':
        assert expr_sym == 'if'
        targs = [translate_expression(lang, arg) for arg in rddl_expr.args]
        return ite(*targs)
    elif expr_type == 'relational':
        tsym = relational_rddl_to_tarski[expr_sym]
        targs = [translate_expression(lang, arg) for arg in rddl_expr.args]
        return create_atom(lang, tsym, targs[0], targs[1])
    elif expr_type == 'aggregation':
        if expr_sym == 'sum':
            var = translate_expression(lang, rddl_expr.args[0])
            sum_expr = translate_expression(lang, rddl_expr.args[1])
            if isinstance(sum_expr, Formula):
                sum_expr = ite(sum_expr, Constant(1, lang.Integer), Constant(0, lang.Integer))
            return tm.summation(var, sum_expr)
        elif expr_sym == 'prod':
            var = translate_expression(lang, rddl_expr.args[0])
            prod_expr = translate_expression(lang, rddl_expr.args[1])
            if isinstance(prod_expr, Formula):
                prod_expr = ite(prod_expr, Constant(1, lang.Integer), Constant(0, lang.Integer))
            return tm.product(var, prod_expr)
        elif expr_sym == 'forall':
            var = translate_expression(lang, rddl_expr.args[0])
            forall_expr = translate_expression(lang, rddl_expr.args[1])
            return forall(var, forall_expr)
    elif expr_type == 'arithmetic':
        op = arithmetic_rddl_to_tarski[expr_sym]
        targs = [lang] + [translate_expression(lang, arg) for arg in rddl_expr.args]
        # MRJ: in some domains (e.g. Mars Rovers) there's stuff like a boolean multiplying an
        # expression
        if expr_sym == '*':
            if isinstance(targs[1], Formula):
                return ite(targs[1], targs[2], Constant(0, targs[2].sort))
            if isinstance(targs[2], Formula):
                return ite(targs[2], targs[1], Constant(0, targs[1].sort))
        # print(expr_sym, targs)
        if len(targs) == 2:  # unary operator
            if expr_sym == '-':
                # replace -x by -1 * x
                actual_op = arithmetic_rddl_to_tarski['*']
                targs = [targs[0]] + [Constant(-1.0, lang.Real)] + [targs[1]]
                return actual_op(*targs)
        return op(*targs)
    elif expr_type == 'func':
        func = func_rddl_to_tarski[expr_sym]
        targs = [translate_expression(lang, arg) for arg in rddl_expr.args]
        return lang.get(func)(*targs)
    elif expr_type == 'randomvar':
        func = func_rddl_to_tarski[expr_sym]
        targs = [translate_expression(lang, arg) for arg in rddl_expr.args]
        return lang.get(func)(*targs)
    elif expr_type == 'boolean':
        connective = logic_rddl_to_tarski[expr_sym]
        targs = [translate_expression(lang, arg) for arg in rddl_expr.args]
        return connective(*targs)
    elif expr_type == 'constant':
        if "class 'float'" in expr_sym:
            k = Constant(rddl_expr.args, lang.Real)
            return k
        if "class 'int'" in expr_sym:
            k = Constant(rddl_expr.args, lang.Real)
            return k

    # print(rddl_expr, expr_type, type(expr_sym))
    assert False


class Parameters:
    """
        RDDL instance parameters
    """

    def __init__(self):
        self.horizon = None
        self.discount = None
        self.max_actions = None


class Reader:
    """
        Reader creates a FOL and several language components
        that specify a RDDL task
    """

    def __init__(self, filename):
        self.language = None
        self.rddl_model = self._load_rddl_model(filename)
        self.parameters = Parameters()
        self.x0 = None

    def _load_rddl_model(self, filename):
        with open(filename, 'r') as input_file:
            rddl = input_file.read()
        parser = RDDLParser()
        parser.build()
        # parse RDDL
        return parser.parse(rddl)

    def _translate_types(self):
        for typename, parent_type in self.rddl_model.domain.types:
            assert parent_type == 'object'
            self.language.sort(typename, self.language.Object)
        non_fluents = self.rddl_model.non_fluents
        if not hasattr(non_fluents, 'objects') or not non_fluents.objects or non_fluents.objects[0] is None:
            return
        for typename, dom in self.rddl_model.non_fluents.objects:
            for o in dom:
                self.language.constant(o, self.language.get(typename))

    def _translate_variables(self):
        non_fluent_terms = self.rddl_model.domain.non_fluents
        for name, term in non_fluent_terms.items():
            translate_variable(self.language, name, term)
        for name, term in self.rddl_model.domain.state_fluents.items():
            translate_variable(self.language, name, term)
        for name, term in self.rddl_model.domain.intermediate_fluents.items():
            translate_variable(self.language, name, term)
        for name, term in self.rddl_model.domain.action_fluents.items():
            translate_variable(self.language, name, term)

    def translate_rddl_model(self):
        # 0. create language
        self.language = language(self.rddl_model.domain.name,
                                 theories=[Theory.EQUALITY, Theory.ARITHMETIC, Theory.SPECIAL, Theory.RANDOM])

        # 1. create types
        self._translate_types()

        # 2. create functions (or predicates) for all variable types
        self._translate_variables()

        # 3. acquire instance parameters
        self.parameters.horizon = self.rddl_model.instance.horizon
        self.parameters.discount = self.rddl_model.instance.discount
        if self.rddl_model.instance.max_nondef_actions != 'pos-inf':
            self.parameters.max_actions = self.rddl_model.instance.max_nondef_actions

        # 4. recover initial state, interpretation of fluents
        self.x0 = Model(self.language)
        self.x0.evaluator = evaluate

        for fluent, value in self.rddl_model.instance.init_state:
            expr = translate_expression(self.language, ('pvar_expr', fluent))
            if isinstance(expr, Term):
                self.x0.setx(expr, value)
            if isinstance(expr, Atom) and value is True:
                self.x0.add(expr.predicate, *expr.subterms)
        for fluent, value in self.rddl_model.non_fluents.init_non_fluent:
            expr = translate_expression(self.language, ('pvar_expr', fluent))
            if isinstance(expr, Term):
                self.x0.setx(expr, value)
            if isinstance(expr, Atom) and value is True:
                self.x0.add(expr.predicate, *expr.subterms)


built_in_type_map = {'object': 'Object', 'real': 'Real', 'int': 'Integer'}
reverse_built_in_type_map = {'Object': 'object', 'Real': 'real', 'Integer': 'int'}


def translate_builtin_type(lang: FirstOrderLanguage, name):
    return lang.get(built_in_type_map[name])


def translate_variable(lang: FirstOrderLanguage, name, term):
    name = name.split('/')[0]  # forget about the arity of the symbol
    params = []
    if term.param_types is not None:
        params = term.param_types
    if term.range in ('bool', 'boolean'):
        signature = tuple(params)
    else:
        signature = tuple(params + [term.range])
    # translate signature to sorts
    t_signature = []
    for s in signature:
        try:
            t_s = lang.get(s)
        except LanguageError:
            t_s = translate_builtin_type(lang, s)
        t_signature += [t_s]
    if term.range in ('bool', 'boolean'):
        return lang.predicate(name, *t_signature)
    return lang.function(name, *t_signature)


class Requirements(Enum):
    CONTINUOUS = "continuous"
    MULTIVALUED = "multivalued"
    REWARD_DET = "reward-deterministic"
    INTERMEDIATE_NODES = "intermediate-nodes"
    PARTIALLY_OBS = "partially-observed"
    CONCURRENT = "concurrent"
    INTEGER_VALUED = "integer-valued"
    CPF_DETERMINISTIC = "cpf-deterministic"

    def __str__(self):
        return self.value.lower()


class PVarTypes(Enum):
    NON_FLUENT = "non-fluent"
    STATE_FLUENT = "state-fluent"
    INTERM_FLUENT = "interm-fluent"
    OBSERV_FLUENT = "observ-fluent"

    def __str__(self):
        return self.value.lower()


class PVarDomains(Enum):
    BOOL = "bool"
    INT = "int"
    REAL = "real"
    OBJECT = "object"
    ENUMERATED = "enumerated"

    def __str__(self):
        return self.value.lower()


class ConstraintType(Enum):
    ACTION = "action"
    STATE = "state"
    INVARIANT = "invariant"

    def __str__(self):
        return self.value.lower()


symbol_map = {
    BPS.EQ: "==",
    BPS.NE: "~=",
    BPS.LT: "<",
    BPS.LE: "<=",
    BPS.GT: ">",
    BPS.GE: ">=",
    Connective.And: "^",
    Connective.Or: "|",
    Connective.Not: "~",
    Quantifier.Exists: "exists",
    Quantifier.Forall: "forall",
    BFS.ADD: "+",
    BFS.SUB: "-",
    BFS.MUL: "*",
    BFS.DIV: "/",
    BFS.POW: "**",
    BFS.MOD: "%"
}

function_map = {
    BFS.MIN: "min",
    BFS.MAX: "max",
    BFS.ABS: "abs",
    BFS.SIN: "sin",
    BFS.COS: "cos",
    BFS.TAN: "tan",
    BFS.ATAN: "atan",
    BFS.ASIN: "asin",
    BFS.EXP: "exp",
    BFS.LOG: "log",
    BFS.ERF: "erf",
    BFS.ERFC: "erfc",
    BFS.SGN: "sgn",
    BFS.SQRT: "sqrt",
    BFS.NORMAL: "Normal",
    BFS.GAMMA: "Gamma",
    BFS.KRON: "KronDelta",
    BFS.DIRAC: "DiracDelta",
    BFS.BERNOULLI: "Bernoulli",
    BFS.DISCRETE: "Discrete",
    BFS.POISSON: "Poisson"
}


class Writer:
    """
        Writer compiles the specification of a planning task
        into RDDL
    """

    def __init__(self, task):
        self.task = task
        self.need_constraints = {}
        self.need_obj_decl = []
        self.non_fluent_signatures = set()
        self.interm_signatures = set()

    def write_model(self, filename):
        tpl = load_tpl("rddl_model.tpl")
        content = tpl.format(
            domain_name=self.task.domain_name,
            req_list=self.get_requirements(),
            type_list=self.get_types(),
            pvar_list=self.get_pvars(),
            cpfs_list=self.get_cpfs(),
            reward_expr=self.get_reward(),
            action_precondition_list=self.get_preconditions(),
            state_invariant_list=self.get_state_invariants(),
            domain_non_fluents='{}_non_fluents'.format(self.task.instance_name),
            object_list=self.get_objects(),
            non_fluent_expr=self.get_non_fluent_init(),
            instance_name=self.task.instance_name,
            init_state_fluent_expr=self.get_state_fluent_init(),
            non_fluents_ref='{}_non_fluents'.format(self.task.instance_name),
            max_nondef_actions=self.get_max_nondef_actions(),
            horizon=self.get_horizon(),
            discount=self.get_discount()
        )
        with open(filename, 'w') as file:
            file.write(content)

    def get_requirements(self):
        return ', '.join([str(r) for r in self.task.requirements])

    def get_types(self):
        from tarski.syntax.sorts import parent
        type_decl_list = []
        for S in self.task.L.sorts:
            if S.builtin or S.name == 'object':
                continue
            if isinstance(S, Interval):
                self.need_constraints[S.name] = S
                continue
            type_decl_list += ['{} : {};'.format(S.name, parent(S).name)]
            self.need_obj_decl += [S]
        return '\n'.join(type_decl_list)

    def get_type(self, fl):
        if isinstance(fl, Atom):
            return 'bool'
        try:
            return reverse_built_in_type_map[fl.symbol.codomain.name]
        except KeyError:
            return fl.symbol.codomain.name

    def get_signature(self, fl):
        if isinstance(fl, CompoundTerm):
            sig = fl.symbol.signature
            head = sig[0]
            domain = sig[1:-1]
        elif isinstance(fl, Atom):
            sig = fl.predicate.signature
            head = sig[0]
            domain = sig[1:]
        else:
            assert False
        if len(domain) == 0:
            return '{}'.format(head)
        return '{}({})'.format(head, ','.join(domain))

    def get_pvars(self):
        pvar_decl_list = []
        # state fluents
        for fl, v in self.task.state_fluents:
            rsig = self.get_signature(fl)
            pvar_decl_list += ['\t{} : {{state-fluent, {}, default = {}}};'.format(rsig, self.get_type(fl), str(v))]
        for fl, level in self.task.interm_fluents:
            rsig = self.get_signature(fl)
            try:
                self.interm_signatures.add(fl.symbol.signature)
            except AttributeError:
                self.interm_signatures.add(fl.predicate.signature)
            pvar_decl_list += ['\t{} : {{interm-fluent, {}, level = {}}};'.format(rsig, self.get_type(fl), str(level))]
        for fl, v in self.task.action_fluents:
            rsig = self.get_signature(fl)
            pvar_decl_list += ['\t{} : {{action-fluent, {}, default = {}}};'.format(rsig, self.get_type(fl), str(v))]
        for fl, v in self.task.non_fluents:
            rsig = self.get_signature(fl)
            try:
                self.non_fluent_signatures.add(fl.symbol.signature)
            except AttributeError:
                self.non_fluent_signatures.add(fl.predicate.signature)
            pvar_decl_list += ['\t{} : {{non-fluent, {}, default = {}}};'.format(rsig, self.get_type(fl), str(v))]
        return '\n'.join(pvar_decl_list)

    def get_cpfs(self):
        cpfs_decl_list = []
        for lhs, rhs in self.task.cpfs:
            cpfs_decl_list += ['\t{} = {};'.format(self.get_fluent(lhs, True), self.rewrite(rhs))]
        return '\n'.join(cpfs_decl_list)

    def get_reward(self):
        return self.rewrite(self.task.reward)

    def get_preconditions(self):
        act_prec_list = []
        for expr, ctype in self.task.constraints:
            if ctype == ConstraintType.ACTION:
                act_prec_list += ['\t{};'.format(self.rewrite(expr))]
        return '\n'.join(act_prec_list)

    def get_state_invariants(self):
        state_inv_list = []
        for expr, ctype in self.task.constraints:
            if ctype == ConstraintType.STATE:
                state_inv_list += ['\t{};'.format(self.rewrite(expr))]
        return '\n'.join(state_inv_list)

    def get_objects(self):
        obj_decl_blocks = []

        if len(self.need_obj_decl) == 0:
            return ''

        # initialize
        for S in self.need_obj_decl:
            domain_str = ','.join([str(c.symbol) for c in S.domain()])
            obj_decl_blocks += ['\t{} : {{{}}};'.format(S.name, domain_str)]

        return 'objects {{{}}};'.format('\n'.join(obj_decl_blocks))

    def get_non_fluent_init(self):
        non_fluent_init_list = []
        for signature, defs in self.task.x0.function_extensions.items():
            if signature not in self.non_fluent_signatures:
                continue
            for st_refs, value in defs.data.items():
                subterms = [r.expr for r in st_refs]
                if len(subterms) == 0:
                    term_str = signature[0]
                else:
                    term_str = str(self.task.L.get(signature[0])(*subterms))
                non_fluent_init_list += ['\t{} = {};'.format(term_str, value)]
        for signature, defs in self.task.x0.predicate_extensions.items():
            if signature not in self.non_fluent_signatures:
                continue
            for st_refs in defs:
                subterms = [r.expr for r in st_refs]
                if len(subterms) == 0:
                    atom_str = signature[0]
                else:
                    atom_str = str(self.task.L.get(signature[0])(*subterms))
                non_fluent_init_list += ['\t{} = true;'.format(atom_str)]

        if len(non_fluent_init_list) == 0:
            return ''

        return 'non-fluents {{{}}};'.format('\n'.join(non_fluent_init_list))

    def get_state_fluent_init(self):
        init_list = []
        for signature, defs in self.task.x0.function_extensions.items():
            if signature in self.non_fluent_signatures \
                    or signature in self.interm_signatures:
                continue
            for st_refs, value in defs.data.items():
                subterms = [r.expr for r in st_refs]
                if len(subterms) == 0:
                    term_str = signature[0]
                else:
                    term_str = str(self.task.L.get(signature[0])(*subterms))
                init_list += ['\t{} = {};'.format(term_str, value)]
        for signature, defs in self.task.x0.predicate_extensions.items():
            if signature in self.non_fluent_signatures \
                    or signature in self.interm_signatures:
                continue
            for st_refs in defs:
                subterms = [r.expr for r in st_refs]
                if len(subterms) == 0:
                    atom_str = signature[0]
                else:
                    atom_str = str(self.task.L.get(signature[0])(*subterms))
                init_list += ['\t{} = true;'.format(atom_str)]

        return '\n'.join(init_list)

    def rewrite(self, expr):
        if expr is None:
            return '0.0'
        if isinstance(expr, CompoundTerm):
            re_st = [self.rewrite(st) for st in expr.subterms]
            if expr.symbol.builtin:
                if expr.symbol.symbol in symbol_map.keys():
                    return '({} {} {})'.format(re_st[0], symbol_map[expr.symbol.symbol], re_st[1])
            st_str = ''
            if expr.symbol.builtin:
                if expr.symbol.symbol in function_map.keys():
                    if len(re_st) > 0:
                        # MRJ: Random variables need parenthesis, other functions need
                        # brackets...
                        if expr.symbol.symbol in builtins.get_random_binary_functions():
                            st_str = '({})'.format(','.join(re_st))
                        else:
                            st_str = '[{}]'.format(','.join(re_st))
                    return '{}{}'.format(function_map[expr.symbol.symbol], st_str)
            if len(re_st) > 0:
                st_str = '({})'.format(','.join(re_st))
            return '{}{}'.format(expr.symbol.signature[0], st_str)
        elif isinstance(expr, Atom):
            re_st = [self.rewrite(st) for st in expr.subterms]
            if expr.predicate.builtin:
                if expr.predicate.symbol in symbol_map.keys():
                    return '({} {} {})'.format(re_st[0], symbol_map[expr.predicate.symbol], re_st[1])
            st_str = ''
            if len(re_st) > 0:
                st_str = '({})'.format(','.join(re_st))
            return '{}{}'.format(expr.predicate.signature[0], st_str)
        elif isinstance(expr, Variable):
            # remove ? just in case
            return '?{}'.format(expr.symbol.replace('?', ''))
        elif isinstance(expr, Constant):
            return str(expr)
        elif isinstance(expr, IfThenElse):
            cond = self.rewrite(expr.condition)
            expr1 = self.rewrite(expr.subterms[0])
            expr2 = self.rewrite(expr.subterms[1])
            return 'if ({}) then ({}) else ({})'.format(cond, expr1, expr2)
        elif isinstance(expr, Tautology):
            return 'true'
        elif isinstance(expr, Contradiction):
            return 'false'
        elif isinstance(expr, CompoundFormula):
            re_sf = [self.rewrite(st) for st in expr.subformulas]
            re_sym = symbol_map[expr.connective]
            if len(re_sf) == 1:
                return '{}{}'.format(re_sym, re_sf)
            return '({} {} {})'.format(re_sf[0], re_sym, re_sf[1])
        elif isinstance(expr, QuantifiedFormula):
            re_f = self.rewrite(expr.formula)
            re_vars = ['?{} : {}'.format(x.symbol, x.sort.name) for x in expr.variables]
            re_sym = symbol_map[expr.quantifier]
            return '{}_{{{}}} ({})'.format(re_sym, ','.join(re_vars), re_f)
        elif isinstance(expr, AggregateCompoundTerm):
            re_expr = self.rewrite(expr.subterm)
            re_vars = ['?{} : {}'.format(x.symbol, x.sort.name) for x in expr.bound_vars]
            if expr.symbol == BFS.ADD:
                re_sym = 'sum'
            elif expr.symbol == BFS.MUL:
                re_sym = 'prod'
            return '{}_{{{}}} ({})'.format(re_sym, ','.join(re_vars), re_expr)

    def get_fluent(self, fl, next_state=False):
        try:
            head = fl.symbol.signature[0]
        except AttributeError:
            head = fl.predicate.signature[0]
        subterms_str = ''
        if len(fl.subterms) > 0:
            subterms_str = []
            for st in fl.subterms:
                if isinstance(st, Variable):
                    # remove any ? from symbol just in case
                    subterms_str += ['?{}'.format(st.symbol.replace('?', ''))]
                else:
                    subterms_str += [str(st)]
            subterms_str = '({})'.format(','.join(subterms_str))
        prima = ''
        if next_state:
            prima = "'"
        return "{}{}{}".format(head, prima, subterms_str)

    def get_max_nondef_actions(self):
        return str(self.task.parameters.max_nondef_actions)

    def get_horizon(self):
        return str(self.task.parameters.horizon)

    def get_discount(self):
        return str(self.task.parameters.discount)




import itertools
import copy

from ...syntax import create_substitution, TermSubstitution
from ... import fstrips as fs
from ...util import IndexDictionary
from . import instantiation
from .elements import process_expression, process_effect


class ActionGrounder:

    def __init__(self, prob, index):
        self.problem = prob
        self.L = self.problem.language
        self.index = index
        self.problem.ground_actions = IndexDictionary()
        self.schemas = list(self.problem.actions.values())
        self.actions_generated = 0

    def __str__(self):
        return 'Actions generated: {}'.format(self.actions_generated)

    def calculate_actions(self):
        for act_schema in self.schemas:
            k, syms, substs = instantiation.enumerate_groundings(self.L, act_schema.parameters)
            for values in itertools.product(*substs):
                subst = create_substitution(syms, values)
                op = TermSubstitution(self.L, subst)
                g_prec = process_expression(self.L, act_schema.precondition, op)
                g_effs = []
                for eff in act_schema.effects:
                    g_eff = copy.deepcopy(eff)
                    g_eff.condition = process_expression(self.L, g_eff.condition, op, False)
                    g_eff = process_effect(self.L, g_eff, op)

                    g_effs.append(g_eff)
                self.problem.ground_actions.add(fs.Action(self.L, act_schema.name, values, g_prec, g_effs))
            self.actions_generated += k



"""

"""
from enum import Enum

from .concepts import Concept, Role, NullaryAtom
from ..utils.algorithms import compute_min_distance
from ..utils.hashing import consistent_hash


class FeatureValueChange(Enum):
    ADD = 1
    DEL = 2
    INC = 3
    DEC = 4
    NIL = 5
    INC_OR_NIL = 6
    ADD_OR_NIL = 7


class Feature:
    def denotation(self, model):
        raise NotImplementedError()

    def diff(self, x, y):
        raise NotImplementedError()

    def __hash__(self):
        raise NotImplementedError()

    def __eq__(self, other):
        raise NotImplementedError()

    @staticmethod
    def bool_value(value):
        assert value >= 0
        return value > 0

    # def concept(self):
    #     raise NotImplementedError()

    def complexity(self):
        """ Return feature complexity value """
        raise NotImplementedError()


def compute_int_feature_diff(x, y):
    assert type(x) is int and x >= 0
    assert type(y) is int and y >= 0
    if x == y:
        return FeatureValueChange.NIL
    if x > y:
        return FeatureValueChange.DEC
    else:
        return FeatureValueChange.INC


def compute_bool_feature_diff(x, y):
    assert type(x) is bool
    assert type(y) is bool
    if y and not x:
        return FeatureValueChange.ADD
    if x and not y:
        return FeatureValueChange.DEL
    else:
        return FeatureValueChange.NIL


def are_feature_changes_analogous(x, y):
    return x == y or \
           (x == FeatureValueChange.DEL and y == FeatureValueChange.DEC) or \
           (x == FeatureValueChange.DEC and y == FeatureValueChange.DEL) or \
           (x == FeatureValueChange.ADD and y == FeatureValueChange.INC) or \
           (x == FeatureValueChange.INC and y == FeatureValueChange.ADD)


class ConceptCardinalityFeature(Feature):
    """ A numeric feature that reflects the cardinality of a set of objects defined by a concept """
    def __init__(self, c):
        assert isinstance(c, Concept)
        self.c = c
        self.hash = consistent_hash((self.__class__, self.c))

    def denotation(self, model):
        return model.compressed_denotation(self.c).count()

    def diff(self, x, y):
        return compute_int_feature_diff(x, y)

    def __repr__(self):
        return 'Num[{}]'.format(self.c)

    __str__ = __repr__

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c == other.c)

    def concept(self):
        return self.c

    def complexity(self):
        # The complexity of a cardinality feature is the complexity of the underlying concept
        return self.c.size


class EmpiricalBinaryConcept(Feature):
    def __init__(self, feature):
        assert isinstance(feature, (ConceptCardinalityFeature, EmpiricalBinaryConcept))
        self.c = feature.c
        self.hash = consistent_hash((self.__class__, self.c))

    def denotation(self, model):
        val = model.compressed_denotation(self.c).count()
        assert val in (0, 1)  # By definition of "empirical binary concept"
        return bool(val)

    def diff(self, x, y):
        return compute_bool_feature_diff(x, y)

    def __repr__(self):
        return 'Bool[{}]'.format(self.c)

    __str__ = __repr__

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c == other.c)

    def concept(self):
        return self.c

    def complexity(self):
        # The complexity of a binary feature is the complexity of the underlying concept
        return self.c.size

# NOT YET FULLY IMPLEMENTED:
#
# class IntegerVariableFeature(Feature):
#     """ A numeric feature that directly reflects the value of some integer variable of the problem """
#     def __init__(self, fun, point):
#         assert isinstance(fun, Function)
#         assert isinstance(point, tuple)
#         self.fun = fun
#         self.point = point
#         self.hash = consistent_hash((self.__class__, self.fun.symbol, point))
#
#     def value(self, cache, state):
#         """ The feature value _is_ the cardinality of the extension of the represented concept"""
#         raise RuntimeError("Unimplemented")
#         # ext = self.c.extension(cache, state)
#         # return ext.count()
#
#     def diff(self, x, y):
#         return compute_int_feature_diff(x, y)
#
#     def __repr__(self):
#         return 'int[{}]'.format(self.fun(*self.point))
#
#     __str__ = __repr__
#
#     def __hash__(self):
#         return self.hash
#
#     def __eq__(self, other):
#         return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
#                 self.fun == other.fun and self.point == other.point)
#


class MinDistanceFeature(Feature):
    def __init__(self, c1, r, c2):
        assert isinstance(c1, Concept) and isinstance(r, Role) and isinstance(c2, Concept)
        self.c1 = c1
        self.r = r
        self.c2 = c2
        self.hash = consistent_hash((self.__class__, self.c1, self.r, self.c2))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c1 == other.c1 and self.r == other.r and self.c2 == other.c2)

    def denotation(self, model):
        """ The value of the feature is the min distance between any object in the extension of c1 and any object
            on the extension of c2, moving only along r-edges.
        """
        ext_c1 = model.uncompressed_denotation(self.c1)
        ext_c2 = model.uncompressed_denotation(self.c2)
        ext_r = model.uncompressed_denotation(self.r)

        # (Debugging)
        # ec1 = sorted(cache.universe.value(x) for x in cache.uncompress(ext_c1, self.c1.ARITY))
        # ec2 = sorted(cache.universe.value(x) for x in cache.uncompress(ext_c2, self.c2.ARITY))
        # er1 = sorted((cache.universe.value(x), cache.universe.va
        # for x, y in cache.uncompress(ext_r, self.r.ARITY))

        return compute_min_distance(ext_c1, ext_r, ext_c2)

    def diff(self, x, y):
        return compute_int_feature_diff(x, y)

    def __repr__(self):
        return 'Dist[{}, {}, {}]'.format(self.c1, self.r, self.c2)

    __str__ = __repr__

    def complexity(self):
        return self.c1.size + self.r.size + self.c2.size + 1


class NullaryAtomFeature(Feature):
    def __init__(self, atom):
        assert isinstance(atom, NullaryAtom)
        self.atom = atom
        self.hash = consistent_hash((self.__class__, self.atom))

    def denotation(self, model):
        """ The feature evaluates to true iff the nullary atom is true in the given state """
        # return self.atom.extension(cache, state)
        return model.primitive_denotation(self.atom)

    def diff(self, x, y):
        return compute_bool_feature_diff(x, y)

    def __repr__(self):
        return 'Atom[{}]'.format(self.atom)

    __str__ = __repr__

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.atom == other.atom)

    def complexity(self):
        return 1




from ..evaluators.simple import evaluate


class SearchModel:
    def init(self):
        raise NotImplementedError()

    def successors(self, state):
        raise NotImplementedError()

    def is_goal(self, state):
        raise NotImplementedError()


class ForwardSearchModel(SearchModel):

    def __init__(self, problem):
        self.problem = problem

    def init(self):
        return self.problem.init

    def successors(self, state):
        # TODO We would need either to have the ground actions or to ground on-the-fly here
        raise NotImplementedError()

    def is_goal(self, state):
        state.evaluator = evaluate  # TODO This is a bit hacky
        return state[self.problem.goal]  # This triggers the logical interpretation of the goal formula



"""
    Universal Quantification Elimination
"""
import itertools

from .subst import create_substitution
from ..formulas import land, Quantifier, QuantifiedFormula
from ..transform import term_substitution, PrenexTransformation
from .errors import TransformationError


class UniversalQuantifierElimination:
    """
        This class rewrites the input formula phi into an equivalent formula
        in Prenex NNF (referenced by the )

    """

    def __init__(self, lang, phi, do_copy=True):
        self.L = lang
        self.blueprint = PrenexTransformation.rewrite(self.L, phi, do_copy).prenex
        # print(str(self.blueprint))
        self.universal_free = None

    def _convert(self, phi):

        if isinstance(phi, QuantifiedFormula):
            if phi.quantifier == Quantifier.Forall:
                card, syms, substs = _enumerate_instantations(phi)
                if card == 0:
                    raise TransformationError("universal elimination", phi, "No constants were defined!")
                conjuncts = []
                for values in itertools.product(*substs):
                    subst = create_substitution(syms, values)
                    conjuncts.append(term_substitution(self.L, phi.formula, subst))
                # print(len(conjuncts))
                return land(*conjuncts)

            else:
                phi.formula = self._convert(phi.formula)
                return phi

        return phi

    def convert(self):
        self.universal_free = self._convert(self.blueprint)

    @staticmethod
    def rewrite(lang, phi, do_copy=True):
        trans = UniversalQuantifierElimination(lang, phi, do_copy)
        trans.convert()
        return trans


def _enumerate_instantations(phi):
    assert isinstance(phi, QuantifiedFormula)
    syms = []
    instantiations = []
    cardinality = 1
    for st in phi.variables:
        if st.sort.builtin:
            raise TransformationError("universal elimination", phi,
                                      "Variable found of built-in sort '{}', domain is too large!".format(
                                          st.sort.name))
        syms.append(st)
        instantiations.append(list(st.sort.domain()))
        cardinality *= len(instantiations[-1])

    return cardinality, syms, instantiations



from .. import errors as err
from .formulas import Atom
from .terms import Term
from .builtins import BuiltinPredicateSymbol, BuiltinFunctionSymbol


def check_same_language(lhs, rhs):
    language = lhs.language
    if language != rhs.language:
        raise err.LanguageMismatch(rhs, rhs.language, language)
    return language


def create_atom(symbol: BuiltinPredicateSymbol, lhs, rhs):
    assert isinstance(lhs, Term) and isinstance(rhs, Term)

    language = check_same_language(lhs, rhs)

    # s1, s2 = lhs.type, rhs.type
    # if language.is_subtype(s1, s2):
    #     return Atom("xxx", [lhs, rhs])

    # TODO AT THE MOMENT WE DO NOT CHECK FOR TYPE SAFETY WITH BUILT-IN TYPES

    predicate = language.get_predicate(symbol)
    return Atom(predicate, [lhs, rhs])


def create_arithmetic_term(symbol: BuiltinFunctionSymbol, lhs, rhs):
    language = check_same_language(lhs, rhs)
    fun = language.get_function(symbol)
    return fun(lhs, rhs)




import hashlib
import sys


def int_to_bytes(value):
    return value.to_bytes((value.bit_length() + 7) // 8, 'big', signed=True) or b'\0'


def consistent_hash(iterable):
    h = hashlib.sha256()
    for elem in iterable:
        if isinstance(elem, str):
            h.update(elem.encode())

        elif isinstance(elem, int):
            h.update(int_to_bytes(elem))

        elif isinstance(elem, type):
            h.update(str(elem).encode())

        elif hasattr(elem, "__hash__"):
            h.update(int_to_bytes(object.__hash__(elem)))

        else:
            h.update(str(elem).encode())
    return int(h.hexdigest(), base=16) & (2**sys.hash_info.width - 1)




import copy

from ... import fstrips as fs
from ...syntax import Variable, symref
from ...syntax.visitors import CollectVariables
from ...syntax.ops import all_variables


def process_expression(language, schema, op, copy_schema=True):
    # pylint: disable=unused-argument
    if copy_schema:
        g_expr = copy.deepcopy(schema)
    else:
        g_expr = schema
    op.visit(g_expr)
    assert len(all_variables(g_expr)) == 0
    return g_expr


def process_effect(language, eff_schema, op):
    # pylint: disable=unused-argument
    if isinstance(eff_schema, (fs.AddEffect, fs.DelEffect)):
        op.visit(eff_schema.atom)
    elif isinstance(eff_schema, fs.LiteralEffect):
        op.visit(eff_schema.lit)
    elif isinstance(eff_schema, fs.FunctionalEffect):
        op.visit(eff_schema.lhs)
        if isinstance(eff_schema.rhs, Variable):
            eff_schema.rhs = op.subst.get(symref(eff_schema.rhs), None)
        else:
            op.visit(eff_schema.rhs)
    elif isinstance(eff_schema, fs.ChoiceEffect):
        op.visit(eff_schema.obj)
        _ = [op.visit(x) for x in eff_schema.variables]

    # MRJ: invariant
    var_collector = CollectVariables()
    if isinstance(eff_schema, (fs.AddEffect, fs.DelEffect)):
        var_collector.visit(eff_schema.atom)
    elif isinstance(eff_schema, fs.FunctionalEffect):
        var_collector.visit(eff_schema.lhs)
        var_collector.visit(eff_schema.rhs)
    elif isinstance(eff_schema, fs.LiteralEffect):
        var_collector.visit(eff_schema.lit)
    elif isinstance(eff_schema, fs.ChoiceEffect):
        var_collector.visit(eff_schema.obj)
        _ = [var_collector.visit(x) for x in eff_schema.variables]
    assert len(var_collector.variables) == 0
    return eff_schema




from .subst import TermSubstitution, term_substitution
from .nnf import NNFTransformation
from .cnf import CNFTransformation
from .prenex import PrenexTransformation
from .univ_elim import UniversalQuantifierElimination
from .neg_builtin import NegatedBuiltinAbsorption



"""
 Classes and methods related to the Logic-Program based grounding  strategy of planning problems.
"""
from ..reachability import create_reachability_lp, run_clingo, parse_model
from ..reachability.asp import SOLVABLE
from .errors import ReachabilityLPUnsolvable
from ..util import IndexDictionary
from .common import StateVariableLite, approximate_symbol_fluency


class LPGroundingStrategy:
    """ An LP problem grounding grounds actions and state variables of a lifted Tarski problem by creating a relaxed
    reachability logic program, solving it with an ASP solver, and parsing the result.
    """
    def __init__(self, problem):
        self.problem = problem
        self.model = None  # We'll cache the solution of the LP here
        self.fluent_symbols, self.static_symbols = approximate_symbol_fluency(problem)

    def ground_state_variables(self):
        """ Create an index all state variables of the problem by exhaustively grounding all predicate and function
        symbols that are considered to be fluent with respect to the problem constants. Thus, if the problem has one
        fluent predicate "p" and one static predicate "q", and constants "a", "b", "c", the result of this operation
        will be the state variables "p(a)", "p(b)" and "p(c)".
        """
        variables = IndexDictionary()
        model = self._solve_lp()

        for symbol in self.fluent_symbols:
            lang = symbol.language
            for binding in model[symbol.symbol]:
                binding_with_constants = tuple(lang.get(c) for c in binding)
                variables.add(StateVariableLite(symbol, binding_with_constants))

        return variables

    def ground_actions(self):
        """  """
        model = self._solve_lp()
        return {k: model[k] for k in self.problem.actions.keys()}

    def iterate_over_schema_groundings(self, schema_name: str):
        """  """
        model = self._solve_lp()
        return (x for x in model[schema_name])

    def _solve_lp(self):
        if self.model is None:
            lp, tr = create_reachability_lp(self.problem)
            model_filename = run_clingo(lp)
            self.model = parse_model(model_filename, tr)

            if len(self.model[SOLVABLE]) != 1:
                raise ReachabilityLPUnsolvable()
        return self.model

    def __str__(self):
        return 'LPGroundingStrategy["{}"]'.format(self.problem.name)

    __repr__ = __str__




from ..fol import FirstOrderLanguage
from ..io import rddl
from ..model import Model
from ..evaluators.simple import evaluate


class Task:

    def __init__(self, lang: FirstOrderLanguage, domain_name: str, instance_name: str):
        self.L = lang
        self.state_fluents = []
        self.interm_fluents = []
        self.action_fluents = []
        self.non_fluents = []
        self.cpfs = []
        self.constraints = []
        self.reward = None
        self.parameters = rddl.Parameters()
        self.domain_name = domain_name
        self.instance_name = instance_name
        self.x0 = Model(self.L)
        self.x0.evaluator = evaluate
        self.requirements = []

    def declare_state_fluent(self, symbol, value):
        self.state_fluents += [(symbol, value)]

    def declare_interm_fluent(self, symbol, level):
        self.interm_fluents += [(symbol, level)]

    def declare_action_fluent(self, symbol, value):
        self.action_fluents += [(symbol, value)]

    def declare_non_fluent(self, symbol, value):
        self.non_fluents += [(symbol, value)]

    def add_constraint(self, expr, ctype):
        self.constraints += [(expr, ctype)]

    def add_cpfs(self, lhs, rhs):
        self.cpfs += [(lhs, rhs)]




from ...syntax import BuiltinFunctionSymbol, CompoundTerm
from . import errors as err


class DifferentialConstraint:
    """ A (possibly lifted) reaction """

    def __init__(self, language, name, parameters, condition, variate, ode):
        self.name = name
        self.language = language
        self.parameters = parameters
        self.condition = condition
        self.variate = variate
        self.ode = ode
        self._check_well_formed()

    def _check_well_formed(self):
        if not isinstance(self.variate, CompoundTerm):
            raise err.InvalidDifferentialConstraintDefinition(self.variate, "Needs to be a compound term")
        if isinstance(self.variate, BuiltinFunctionSymbol):
            raise err.InvalidDifferentialConstraintDefinition(self.variate, "Cannot be a built-in function")
        # ....

    def ident(self):
        params = ', '.join([str(o) for o in self.parameters])
        return '{}({})'.format(self.name, params)

    def dump(self):
        return dict(name=self.name,
                    params=[par.dump() for par in self.parameters],
                    condition=self.condition.dump(),
                    variate=self.variate.dump(),
                    ode=self.ode.dump())

    def __str__(self):
        tokens = ['reaction {}:'.format(self.name),
                  'cond: ({})'.format(self.condition),
                  'variate: {}'.format(self.variate),
                  'ode: {}'.format(self.ode)]
        return '\n'.join(tokens)



"""
 Classes and methods related to the naive grounding strategy of planning problems.
"""
import itertools

from ..syntax import Constant, Variable, CompoundTerm, Atom, create_substitution, term_substitution,\
    termlists_are_equal, termlist_hash
from ..errors import DuplicateDefinition
from .errors import UnableToGroundError
from .common import approximate_symbol_fluency, StateVariableLite
from ..util import IndexDictionary
from ..fstrips.visitors import FluentSymbolCollector, FluentHeuristic


class ProblemGrounding:
    """ A ProblemGrounding contains information about the grounding of a lifted Tarski problem, possibly including
    information about ground state variables and ground actions.
    """
    def __init__(self, problem):
        self.problem = problem
        self.static_terms = None
        self.fluent_terms = None
        self.state_variables = None

    def _check_static_not_fluents(self):
        """
            Sorts fluent and static sets, so that the only
            static expressions are those which haven't been flagged
            as fluent by at least one of our heuristics.
        """
        self.static_terms = {x for x in self.static_terms if x not in self.fluent_terms}
        assert all(x not in self.static_terms for x in self.fluent_terms)

    def process_symbols(self, problem):
        lang = problem.language
        self.fluent_terms = set()
        self.static_terms = set()

        prec_visitor = FluentSymbolCollector(lang, self.fluent_terms, self.static_terms, FluentHeuristic.precondition)
        eff_visitor = FluentSymbolCollector(lang, self.fluent_terms, self.static_terms, FluentHeuristic.action_effects)
        constr_visitor = FluentSymbolCollector(lang, self.fluent_terms, self.static_terms, FluentHeuristic.constraint)

        o_f = len(self.fluent_terms)
        o_s = len(self.static_terms)
        while True:
            problem.get_symbols(prec_visitor, eff_visitor, constr_visitor)
            self._check_static_not_fluents()
            if len(self.fluent_terms) == o_f and len(self.static_terms) == o_s:
                break
            o_f = len(self.fluent_terms)
            o_s = len(self.static_terms)

    def compute_fluent_and_statics(self):
        """ Return sets with fluent and static predicate / function symbols """
        fluents = set(ref.expr.predicate for ref in self.fluent_terms)
        statics = set(ref.expr.predicate for ref in self.static_terms)
        statics = set(x for x in statics if x not in fluents and not x.builtin)
        return fluents, statics


def create_all_possible_state_variables(fluent_terms):
    """ Creates an index with all possible state variables by brute-force
        enumeration.
    """
    variables = IndexDictionary()

    for ref in fluent_terms:
        # @TODO: Work in Progress and we will need to iterate over this a bit
        # @TODO: Sort fluent symbols according to the number of variable subterms
        # L = ref.language
        instantiations = []
        for st in ref.expr.subterms:
            if isinstance(st, Constant):
                instantiations.append([st])
            elif isinstance(st, Variable):
                if st.sort.builtin:
                    raise UnableToGroundError(st, "Term is of built-in sort '{}', domain is too large!".format(
                        st.sort.name))
                instantiations.append(list(st.sort.domain()))
            else:
                raise UnableToGroundError(st, "Grounding of complex nested subterms is not implemented yet!")
        for instantiation in itertools.product(*instantiations):
            try:
                variables.add(StateVariable(ref.expr, instantiation))
            except DuplicateDefinition:
                pass
    return variables


class StateVariable:
    """ A state variable is nothing else than a CompoundTerm or Atom which is expected to change its
    value, along with a particular instantiation of its subterms.
    """

    def __init__(self, term, instantiation):
        assert isinstance(term, (CompoundTerm, Atom))
        self.term = term
        self.head = term.predicate if isinstance(term, Atom) else term.symbol
        self.instantiation = instantiation

    def __hash__(self):
        return hash((self.head.symbol, termlist_hash(self.instantiation)))

    def __eq__(self, other):
        return self.head.symbol == other.head.symbol and termlists_are_equal(self.instantiation, other.instantiation)

    def __str__(self):
        return '{}({})'.format(self.head.symbol, ','.join(str(a) for a in self.instantiation))

    __repr__ = __str__

    @property
    def ground(self):
        subst = create_substitution(self.term.subterms, self.instantiation)
        return term_substitution(self.head.language, self.term, subst)


class NaiveGroundingStrategy:
    """ A naive problem grounding grounds actions and state variables of a lifted Tarski problem by (type-informed)
    exhaustive enumeration of all possible subsitutions of the representation variables.
    Note: This is a lightweight version of the ProblemGrounding class above, hoping that it can eventually replace it.
    """
    def __init__(self, problem):
        self.problem = problem
        self.fluent_symbols, self.static_symbols = approximate_symbol_fluency(problem)

    def ground_state_variables(self):
        """ Create an index all state variables of the problem by exhaustively grounding all predicate and function
        symbols that are considered to be fluent with respect to the problem constants. Thus, if the problem has one
        fluent predicate "p" and one static predicate "q", and constants "a", "b", "c", the result of this operation
        will be the state variables "p(a)", "p(b)" and "p(c)".
        """
        return ground_symbols_exhaustively(self.fluent_symbols)

    def ground_actions(self):
        """ """
        raise NotImplementedError()

    def __str__(self):
        return 'NaiveGroundingStrategy["{}"]'.format(self.problem.name)

    __repr__ = __str__


def ground_symbols_exhaustively(symbols):
    """ Creates an index with all possible groundings of the given predicate and function symbols
    in the given language """
    variables = IndexDictionary()

    for symbol in symbols:
        # We need to consider full sort for predicates, domain only for functions
        domains = [s.domain() for s in symbol.domain]

        for binding in itertools.product(*domains):
            variables.add(StateVariableLite(symbol, binding))

    return variables




import itertools

from ...syntax import QuantifiedFormula, Quantifier, create_substitution
from ...syntax.transform import TermSubstitution, UniversalQuantifierElimination, NegatedBuiltinAbsorption, \
    CNFTransformation
from ...syntax.ops import all_variables
from ...util import IndexDictionary
from . import instantiation
from .elements import process_expression


class ConstraintGrounder:

    def __init__(self, prob, index):
        self.problem = prob
        self.L = self.problem.language
        self.index = index
        self.problem.ground_constraints = IndexDictionary()
        self.schemas = list(self.problem.constraints)
        self.constraints_generated = 0

    def __str__(self):
        return 'Constraints Generated: {}'.format(self.constraints_generated)

    def calculate_constraints(self):

        for const_schema in self.schemas:
            # 1. Collect set of free variables in the constraint
            const_schema = UniversalQuantifierElimination.rewrite(self.L, const_schema).universal_free
            K, syms, substs = instantiation.enumerate_groundings(self.L, all_variables(const_schema))
            for values in itertools.product(*substs):
                subst = create_substitution(syms, values)
                op = TermSubstitution(self.L, subst)
                g_const = process_expression(self.L, const_schema, op)
                # Simplification steps
                s0 = NegatedBuiltinAbsorption.rewrite(self.L, g_const)
                # CNF
                if isinstance(s0.formula, QuantifiedFormula):
                    assert s0.formula.quantifier == Quantifier.Exists
                    # s0 is of the form \exists x1, ..., xn phi, phi is quantifier free
                    s0.formula.formula = CNFTransformation.rewrite(self.L, s0.formula.formula).cnf
                    self.problem.ground_constraints.add(s0.formula)
                else:
                    s1 = CNFTransformation.rewrite(self.L, s0.formula)
                    self.problem.ground_constraints.add(s1.cnf)
            self.constraints_generated += K




class TarskiError(Exception):
    """ Common ancestor class to all of Tarski's exceptions """


class LanguageError(TarskiError):
    def __init__(self, msg=None):
        msg = msg or 'Unexplained Tarski language error'
        super().__init__(msg)


class SyntacticError(LanguageError):
    def __init__(self, msg=None):
        msg = msg or 'Unexplained Tarski syntactic error'
        super().__init__(msg)


class SemanticError(LanguageError):
    def __init__(self, msg=None):
        msg = msg or 'Unexplained Tarski semantic error'
        super().__init__(msg)


class LanguageMismatch(SyntacticError):
    def __init__(self, obj, l1, l2, msg=None):
        msg = msg or ('Language mismatch when operating on object {obj} of type {classname}.\n'
                      'Expected language: {l2}\n'
                      'Actual language: : {l1}\n') \
            .format(obj=obj, classname=type(obj).__name__, l1=l1, l2=l2)
        super().__init__(msg)


class ArityMismatch(SyntacticError):
    def __init__(self, head, arguments, msg=None):
        msg = msg or 'Arity mismatch applying element {} with arity {} to arguments {}'. \
            format(head, head.arity, arguments)
        super().__init__(msg)


class SortMismatch(SyntacticError):
    def __init__(self, element, type_, expected_type, msg=None):
        msg = msg or 'Sort mismatch on element {}. Expected sort was "{}", element has sort "{}"'.format(
            element, expected_type, type_)
        super().__init__(msg)


class DuplicateDefinition(SyntacticError):
    def __init__(self, name, other, msg=None):
        msg = msg or 'Duplicate definition of element "{}": "{}"'.format(name, other)
        super().__init__(msg)
        self.name = name
        self.other = other


class UndefinedElement(SyntacticError):
    def __init__(self, name, msg=None):
        if msg is not None:
            msg = 'Undefined element {}, {}'.format(name, msg)
        else:
            msg = 'Undefined element: {}'.format(name)
        super().__init__(msg)


class DuplicateSortDefinition(DuplicateDefinition):
    pass


class DuplicatePredicateDefinition(DuplicateDefinition):
    pass


class DuplicateFunctionDefinition(DuplicateDefinition):
    pass


class DuplicateConstantDefinition(DuplicateDefinition):
    pass


class DuplicateActionDefinition(DuplicateDefinition):
    pass


class DuplicateVariableDefinition(DuplicateDefinition):
    def __init__(self, variable, other, msg=None):
        msg = msg or "Variable with name '{}' already defined in binding: {}".format(variable.symbol, other)
        super().__init__(variable, other, msg)


class UndefinedTerm(UndefinedElement):
    pass


class UndefinedSort(UndefinedElement):
    pass


class UndefinedPredicate(UndefinedElement):
    pass


class UndefinedFunction(UndefinedElement):
    pass


class UndefinedConstant(UndefinedElement):
    pass


class UndefinedAction(UndefinedElement):
    pass


class UndefinedVariable(UndefinedElement):
    pass


class UnboundVariable(SemanticError):
    def __init__(self, var, msg=None):
        msg = msg or 'Attempted to evaluate open formula with free variable {}'.format(var)
        super().__init__(msg)


class IncorrectExtensionDefinition(SemanticError):
    def __init__(self, element, point, value, msg=None):
        msg = msg or 'Incorrect definition of extension of symbol "{}". Cannot assign value "{}" to point "{}"'.format(
            element, value, point)
        super().__init__(msg)


class UnknownTheory(LanguageError):
    def __init__(self, theory):
        super().__init__('Unknown first-order theory "{}"'.format(theory))


class CommandNotFoundError(TarskiError):
    def __init__(self, name, msg=None):
        msg = msg or 'Necessary command "{}" could not be found'.format(name)
        super().__init__(msg)


class ExternalCommandError(TarskiError):
    pass


# class WrongTermUsageError(TarskiError):
#     def __init__(self):
#         super().__init__("""Tarski terms overload the equality operator `__eq__` to allow the
#         construction of FOL atoms such as "loc(b1)==table". A side-effect of this is that Term objects cannot be
#         inserted as such into associative containers such as dictionaries or sets. In order to use those, you will
#         need to wrap the Term object with a call to `symref`, as e.g. in counter[symref(c)] = 2
#         """)



import sys
from collections import deque, defaultdict


def transitive_closure(elements):
    closure = set(elements)
    while True:
        closure_until_now = closure | set((x, w) for x, y in closure for q, w in closure if q == y)

        if len(closure_until_now) == len(closure):
            break

        closure = closure_until_now

    return closure


def compute_min_distance(c1s, relation, c2s):
    """  """
    # Cover first a couple of base cases to enhance performance
    if c1s & c2s:
        return 0
    if not c1s or not c2s or not relation:
        return sys.maxsize

    # index the adjacency relation
    adjacencies = defaultdict(list)
    for s, t in relation:
        adjacencies[s].append(t)

    # initialize table of distances
    min_distances = defaultdict(lambda: sys.maxsize)
    queue = deque()
    for obj in c1s:
        queue.appendleft(obj)
        min_distances[obj] = 0

    # Perform breadth-first search
    while queue:
        s = queue.pop()
        dist_through_s = min_distances[s]+1
        for t in adjacencies[s]:
            if dist_through_s < min_distances[t]:
                min_distances[t] = min_distances[s]+1
                queue.appendleft(t)

    # Return the minimum distance
    return min(min_distances[o] for o in c2s)




from collections import OrderedDict

from .. import model
from .action import Action
from .derived import Derived
from . import fstrips as fs
from . import errors as err


class Problem:
    """ A Functional STRIPS problem """

    def __init__(self, problem_name=None, domain_name=None):
        self.name = problem_name
        self.domain_name = domain_name
        self.language = None
        self.init = None
        self.goal = None
        self.constraints = []
        self.actions = OrderedDict()
        self.derived_ = OrderedDict()
        self.metric_ = None

    @property
    def derived_predicates(self):
        return self.derived_

    @property
    def plan_metric(self):
        return self.metric_

    def action(self, name, parameters, precondition, effects):
        if name in self.actions:
            raise err.DuplicateActionDefinition(name, self.actions[name])

        self.actions[name] = Action(self.language, name, parameters, precondition, effects)
        return self.actions[name]

    def derived(self, name, parameters, formula):
        if name in self.derived_:
            raise err.DuplicateDerivedDefinition(name, self.derived_[name])

        self.derived_[name] = Derived(self.language, name, parameters, formula)
        return self.derived_[name]

    def has_action(self, name):
        return name in self.actions

    def get_action(self, name):
        if not self.has_action(name):
            raise err.UndefinedAction(name)
        return self.actions[name]

    def get_symbols(self, pv, ev, cv):
        """
            Applies visitors to syntactic elements matching the definitions for
            preconditions, effects and constraints.

            parameters:
                - pv: FluentSymbolCollector tuned for preconditions
                - ev: FluentSymbolCollector tuned for effects
                - cv: FluentSymbolCollector tuned for constraints
        """
        for _, act in self.actions.items():
            pv.visit(act.precondition)
            for eff in act.effects:
                pv.visit(eff.condition)
                if isinstance(eff, (fs.AddEffect, fs.DelEffect)):
                    ev.visit(eff.atom)
                elif isinstance(eff, fs.LiteralEffect):
                    ev.visit(eff.lit)
                elif isinstance(eff, fs.FunctionalEffect):
                    ev.visit(eff.lhs)
                elif isinstance(eff, fs.ChoiceEffect):
                    ev.visit(eff.obj)
                    _ = [ev.visit(x) for x in eff.variables]
                else:
                    raise RuntimeError("Effect type '{}' cannot be analysed".format(type(eff)))

        for const in self.constraints:
            cv.reset()
            cv.visit(const)

    def metric(self, opt_expression, opt_type):
        self.metric_ = fs.OptimizationMetric(opt_expression, opt_type)

    def __str__(self):
        return 'FSTRIPS Problem "{}", domain "{}"'.format(self.name, self.domain_name)

    __repr__ = __str__


def create_fstrips_problem(language, problem_name=None, domain_name=None):
    """ Creates an FSTRIPS problem with empty initial state and with no assigned goal """
    problem_name = problem_name or "Unnamed FSTRIPS problem"
    domain_name = domain_name or "Unnamed FSTRIPS domain"
    problem = Problem(problem_name=problem_name, domain_name=domain_name)
    problem.language = language
    problem.init = model.create(language)
    return problem




import itertools

from .. import errors as err
from .terms import CompoundTerm
from .sorts import Sort


class Function:
    def __init__(self, symbol, language, *args):
        self.symbol = symbol
        self.language = language
        self.domain = tuple(args[:-1])
        self.codomain = args[-1]
        self.builtin = False

        self._check_well_formed()

    def _check_well_formed(self):

        for k, a in enumerate(itertools.chain(self.domain, [self.codomain])):
            if not isinstance(a, Sort):
                raise err.LanguageError("Function: argument #{} ('{}') is of type '{}' instead of Sort"
                                        .format(k + 1, a, type(a)))

            if self.language != a.language:
                raise err.LanguageMismatch(a, a.language, self.language)

    @property
    def signature(self):
        return tuple([self.symbol] + [a.name for a in self.domain] + [self.codomain.name])

    @property
    def arity(self):
        return len(self.domain)

    @property
    def sort(self):
        return self.domain + (self.codomain, )

    def __hash__(self):
        return hash(self.signature)

    def dump(self):
        return dict(symbol=self.symbol,
                    domain=[a.name for a in self.domain],
                    codomain=self.codomain.name)

    def uniform_arity(self):
        return len(self.domain) + 1

    def __eq__(self, other):
        return self.signature == other.signature

    def __call__(self, *args):
        return CompoundTerm(self, args)

    def __str__(self):
        return "{}/{}".format(self.symbol, self.arity)

    __repr__ = __str__

    # def __str__(self):
    #     return '{}({})'.format(self.symbol, ','.join([a.name for a in self.domain]))




from ..errors import TarskiError, DuplicateDefinition, UndefinedElement


class IncompleteProblemError(TarskiError):
    def __init__(self, problem, msg=None):
        msg = msg or 'specification is incomplete!'
        super().__init__('Problem "{}": {}'.format(problem.name, msg))


class InvalidEffectError(TarskiError):
    def __init__(self, effect, msg=None):
        msg = msg or 'definition of effect "{}" is invalid!'.format(effect.tostring())
        super().__init__('{}'.format(msg))


class DuplicateActionDefinition(DuplicateDefinition):
    pass


class DuplicateDerivedDefinition(DuplicateDefinition):
    pass


class UndefinedAction(UndefinedElement):
    pass


class InvalidDerivedPredicateError(TarskiError):
    def __init__(self, symbol, formula, msg=None):
        if msg is None:
            msg = ' '

        msg = 'definition of derived predicate "{} \\equiv {}" is invalid! {}'.format(symbol, formula, msg)
        super().__init__(msg)




from ..errors import LanguageError
from .sorts import Sort
from .terms import Constant


class Predicate:
    def __init__(self, name, language, *args):

        self._symbol = name  # name, a string
        self.language = language  # fol the symbol belongs to
        self.sort = []
        self.builtin = False

        # we validate the arguments now
        for k, a in enumerate(args):
            if not isinstance(a, Sort):
                raise LanguageError("Predicate.__init__() : arguments need \
                to be of type 'Sort', {}-th argument '{}' is of type '{}''".format(k + 1, a, type(a)))
            if self.language != a.language:
                raise LanguageError("Predicate.__init__(): {}-th argument \
                belongs to a different language".format(k + 1))
            self.sort.append(a)

    @property
    def symbol(self):
        return self._symbol

    @property
    def signature(self):
        return tuple([self.symbol] + [a.name for a in self.sort])

    @property
    def arity(self):
        return len(self.sort)

    def uniform_arity(self):
        return len(self.sort)

    @property
    def domain(self):
        return self.sort

    # def __str__(self):
    #     return '{}({})'.format(self.symbol, ','.join([a.name for a in self.sort]))

    def __hash__(self):
        return hash(self.signature)

    def __eq__(self, other):
        return self.signature == other.signature

    def __str__(self):
        return "{}/{}".format(self.symbol, self.arity)

    __repr__ = __str__

    def dump(self):
        return dict(symbol=self.symbol, sort=[a.name for a in self.sort])

    def __call__(self, *args):
        from .formulas import Atom
        return Atom(self, args)

    def check_arguments(self, *args):
        if len(args) != self.arity:
            raise LanguageError('Error setting predicate  extension: arity of {} is {}, {} arguments given'.
                                format(self.symbol, self.arity, len(args)))

        for k, arg in enumerate(args):
            if not isinstance(arg, Constant):
                raise LanguageError('Error setting predicate extension: {}-th argument is not a constant'.format(k + 1))
            if arg.sort.name != self.sort[k].name:
                raise LanguageError('Error setting predicate extension: {}-th argument type mismatch, type is {}, '
                                    'was expected to be {}'.format(k + 1, arg.sort.name, self.sort[k].name))

    def satisfied(self, *args):
        # pylint: disable=unused-argument
        raise ValueError("Extensionally defined")




from ..errors import TarskiError


class UnableToGroundError(TarskiError):
    def __init__(self, sym, msg=None):
        msg = msg or 'Reason unspecified'
        super().__init__('Unable to ground Term/Atom "{}": {}'.format(sym, msg))


class ReachabilityLPUnsolvable(TarskiError):
    def __init__(self, msg=None):
        msg = msg or 'The relaxed-reachabilty logic program is not solvable'
        super().__init__(msg)




import itertools
import copy

from ...fstrips import hybrid
from ...syntax import create_substitution, TermSubstitution
from ...util import IndexDictionary
from . import instantiation
from .elements import process_expression, process_effect


class ReactionGrounder:

    def __init__(self, prob, index):
        self.problem = prob
        self.L = self.problem.language
        self.index = index
        self.problem.ground_reactions = IndexDictionary()
        self.schemas = list(self.problem.reactions.values())
        self.reactions_generated = 0

    def __str__(self):
        return 'Reactions generated: {}'.format(self.reactions_generated)

    def calculate_reactions(self):

        for react_schema in self.schemas:
            if len(react_schema.parameters) == 0:
                self.problem.ground_reactions.add(hybrid.Reaction(self.L,
                                                                  react_schema.name, [], react_schema.condition,
                                                                  react_schema.effect))
                self.reactions_generated += 1
                continue

            k, syms, substs = instantiation.enumerate_groundings(self.L, react_schema.parameters)
            for values in itertools.product(*substs):
                subst = create_substitution(syms, values)

                op = TermSubstitution(self.L, subst)

                g_cond = process_expression(self.L, react_schema.condition, op)

                g_eff = copy.deepcopy(react_schema.effect)
                g_eff.condition = process_expression(self.L, g_eff.condition, op, False)
                g_eff = process_effect(self.L, g_eff, op)

                self.problem.ground_reactions.add(hybrid.Reaction(self.L, react_schema.name, [], g_cond, g_eff))
            self.reactions_generated += k




import copy
import itertools
from collections import defaultdict, OrderedDict

from . import errors as err
from .syntax import Function, Constant, Variable, Sort, inclusion_closure, Predicate, Interval, sorts
from .syntax.algebra import Matrix


def language(name='L'):
    """ A helper to construct languages"""
    lang = FirstOrderLanguage(name)
    return lang


class FirstOrderLanguage:
    """ A full-fledged many-sorted first-order language """

    def __init__(self, name='L'):
        self.name = name
        self._sorts = {}

        # MRJ: let's represent this temporally as pairs of names of sorts,
        # lhs \sqsubseteq rhs, lhs is a subset of rhs
        self.immediate_parent = dict()

        # ancestor_sorts[t] is a set containing all supertypes of sort 't', but NOT 't'
        self.ancestor_sorts = defaultdict(set)

        self._functions = {}
        self._predicates = {}
        # self._predicates_by_sort = {}
        # self._functions_by_sort = {}
        self._constants = OrderedDict()

        self._operators = dict()
        self._global_index = dict()
        self._element_containers = {Sort: self._sorts,
                                    Function: self._functions,
                                    Predicate: self._predicates}

        self.language_components_frozen = False
        self.theories = []

        self._build_builtin_sorts()

    def __deepcopy__(self, memo):
        """ At the moment we forbid deep copies of this class, as they might be too expensive"""
        memo[id(self)] = self
        return self

    def deepcopy(self):
        """ Use this method instead of copy.deepcopy() if you need a true deep-copy of the language """
        memo = dict()
        newone = type(self)()
        memo[id(self)] = newone
        for k, v in self.__dict__.items():
            setattr(newone, k, copy.deepcopy(v, memo))
        return newone

    @property
    def sorts(self):
        for s in self._sorts.values():
            yield s

    @property
    def predicates(self):
        return self._predicates.values()

    @property
    def functions(self):
        return self._functions.values()

    @property
    def Object(self):
        return self._sorts['object']

    @property
    def Real(self):
        return self._sorts['Real']

    @property
    def Integer(self):
        return self._sorts['Integer']

    @property
    def Natural(self):
        return self._sorts['Natural']

    def _build_builtin_sorts(self):
        self._build_the_objects()
        self._build_the_reals()
        self._build_the_integers()
        self._build_the_naturals()

    def _build_the_reals(self):
        the_reals = sorts.build_the_reals(self)
        self._sorts['Real'] = the_reals
        self.set_parent(the_reals, self.Object)
        # self.create_builtin_predicates(the_reals)
        self._global_index['Real'] = the_reals

    def _build_the_integers(self):
        the_ints = sorts.build_the_integers(self)
        self._sorts['Integer'] = the_ints
        self.set_parent(the_ints, self.Real)
        # self.create_builtin_predicates(the_ints)
        self._global_index['Integer'] = the_ints

    def _build_the_naturals(self):
        the_nats = sorts.build_the_naturals(self)
        self._sorts['Natural'] = the_nats
        self.set_parent(the_nats, self.Integer)
        # self.create_builtin_predicates(the_nats)
        self._global_index['Natural'] = the_nats

    def _build_the_objects(self):
        sort = Sort('object', self)
        self._sorts['object'] = sort
        self.immediate_parent[sort] = None
        self.ancestor_sorts[sort] = set()

    def sort(self, name: str, parent: Sort = None):
        """
            Create new sort with given name and parent sort.
            If no parent is specified, the topmost "object" sort is assumed as parent.

            Raises err.DuplicateSortDefinition if a sort with the same name already existed,
             or err.DuplicateDefinition if some non-sort element with the same name already existed.
        """
        self._check_name_not_defined(name, self._sorts, err.DuplicateSortDefinition)

        sort = Sort(name, self)
        self._sorts[name] = sort
        self._global_index[name] = sort

        parent = parent or self.get_sort("object")
        self.set_parent(sort, parent)

        # self.create_builtin_predicates(sort)
        return sort

    def has_sort(self, name):
        return name in self._sorts

    def get_sort(self, name):
        if not self.has_sort(name):
            raise err.UndefinedSort(name)
        return self._sorts[name]

    def interval(self, name, parent: Interval, lower_bound=None, upper_bound=None):
        """ Create a (bound) interval sort.

        We allow only the new sort to derive from the built-in natural, integer or real sorts.
        """
        self._check_name_not_defined(name, self._sorts, err.DuplicateSortDefinition)

        if parent not in (self.Real, self.Natural, self.Integer):
            raise err.SemanticError("Cannot create interval that does not subclass one of "
                                    "the real, integer or natural sort")

        if upper_bound <= lower_bound:
            raise err.SemanticError("Cannot create interval where the upper bound is greater or "
                                    "equal than the lower bound")

        sort = Interval(name, self, parent.encode, lower_bound, upper_bound)
        sort.builtin = parent.builtin
        self._sorts[name] = sort
        self._global_index[name] = sort

        self.set_parent(sort, parent)

        return sort

    def variable(self, name: str, sort: Sort):
        sort = self._retrieve_object(sort, Sort)
        return Variable(name, sort)

    def set_parent(self, sort: Sort, parent: Sort):
        if parent.language is not self:
            raise err.LanguageError("Tried to set as parent a sort from a different language")

        p = self.immediate_parent.get(sort, None)
        if p is not None:
            raise err.LanguageError('Tried to set parent of sort "{}", which has already parent {}'.format(sort, p))

        self.immediate_parent[sort] = parent
        self.ancestor_sorts[sort].update(inclusion_closure(parent))

    def _retrieve_object(self, obj, type_):
        """
        Make sure that the given obj is either an object of a certain language type (e.g. sort, predicate, etc.)
        which has been correctly registered with the language, or the name of such an object, and return the object
        """
        if not isinstance(obj, (str, type_)):
            raise err.LanguageError('Unknown type of language element "{}"'.format(obj))

        if isinstance(obj, type_):
            if obj.language != self:
                raise err.LanguageMismatch(obj, obj.language, self)
            return obj

        # obj must be a string, which we take as the name of a language element
        if type_ not in self._element_containers:
            raise RuntimeError("Trying to index incorrect type {}".format(type_))

        if obj not in self._element_containers[type_]:
            raise err.UndefinedElement(obj)

        return self._element_containers[type_][obj]

    def constant(self, name, sort: Sort):
        """ Create constant symbol of a given sort """
        sort = self._retrieve_object(sort, Sort)

        if sort.builtin:
            if sort.cast(name) is None:
                raise err.SemanticError("Cannot create constant with sort '{}' from '{}' of Python type '{}'".
                                        format(sort.name, name, type(name)))

            # MRJ: if name is a Python primitive type literal that can be interpreted as the underlying
            # type of the built in sort, we return a Constant object.
            # TODO: I don't see it is desirable to store constants of built in sorts.
            # MRJ: We're not storing them anywhere, but we need the Python literals
            # to be decorated so we can use them in our ASTs
            return Constant(name, sort)

        self._check_name_not_defined(name, self._constants, err.DuplicateConstantDefinition)

        c = Constant(name, sort)
        self._constants[name] = c
        self._global_index[name] = c
        return c

    def has_constant(self, name):
        return name in self._constants

    def get_constant(self, name):
        if not self.has_constant(name):
            raise err.UndefinedConstant(name)
        return self._constants[name]

    def constants(self):
        return list(self._constants.values())

    def vector(self, arraylike, sort: Sort):
        import numpy as np
        return Matrix(np.reshape(arraylike, (len(arraylike), 1)), sort)

    def matrix(self, arraylike, sort: Sort):
        return Matrix(arraylike, sort)

    def _check_name_not_defined(self, name, where, exception):
        """ Check whether the given name is already defined in the given container, raising an exception of
        the indicated type in case it is. If not, check it is not defined in any other container, raising a
        more generic DuplicateDefinition exception if it is. """
        if name in where:
            raise exception(name, where[name])

        if name in self._global_index:
            raise err.DuplicateDefinition(name, self._global_index[name])

    def predicate(self, name: str, *args):
        self._check_name_not_defined(name, self._predicates, err.DuplicatePredicateDefinition)

        types = [self._retrieve_object(a, Sort) for a in args]  # Convert possible strings into Sort objects
        predicate = Predicate(name, self, *types)
        self._predicates[name] = predicate
        self._global_index[name] = predicate
        # self._predicates_by_sort[(name,) + tuple(*args)] = predicate
        return predicate

    def has_predicate(self, name):
        return name in self._predicates

    def get_predicate(self, name):
        if not self.has_predicate(name):
            raise err.UndefinedPredicate(name)
        return self._predicates[name]

    def function(self, name: str, *args):
        self._check_name_not_defined(name, self._functions, err.DuplicateFunctionDefinition)

        types = [self._retrieve_object(a, Sort) for a in args]  # Convert possible strings into Sort objects
        func = Function(name, self, *types)
        self._functions[name] = func
        self._global_index[name] = func
        # self._functions_by_sort[(name,) + tuple(*args)] = func
        return func

    def has_function(self, name):
        return name in self._functions

    def get_function(self, name):
        if not self.has_function(name):
            raise err.UndefinedFunction(name)
        return self._functions[name]

    def dump(self):
        return dict(
            sorts=[s.dump() for _, s in self._sorts.items()],
            predicates=[p.dump() for _, p in self._predicates.items()],
            functions=[f.dump() for _, f in self._functions.items()]
        )

    def check_well_formed(self):
        for _, s in self._sorts.items():
            if s.cardinality() == 0:
                raise err.LanguageError("Sort '{}' is empty!".format(s))

    def most_restricted_type(self, t1, t2):
        if self.is_subtype(t1, t2):
            return t1
        elif self.is_subtype(t2, t1):
            return t2
        return None

    def is_subtype(self, t, st):
        t = self._retrieve_object(t, Sort)
        st = self._retrieve_object(st, Sort)
        return t == st or self.is_strict_subtype(t, st)

    def is_strict_subtype(self, t, st):
        t = self._retrieve_object(t, Sort)
        st = self._retrieve_object(st, Sort)
        return st in self.ancestor_sorts[t]

    def are_vertically_related(self, t1, t2):
        t1 = self._retrieve_object(t1, Sort)
        t2 = self._retrieve_object(t2, Sort)
        return self.is_subtype(t1, t2) or self.is_subtype(t2, t1)

    def __str__(self):
        return "{}: Tarski language with {} sorts, {} function symbols, {} predicate symbols".format(
            self.name, len(self._sorts), len(self._functions), len(self._predicates))

    def register_operator_handler(self, operator, t1, t2, handler):
        self._operators[(operator, t1, t2)] = handler

    def register_unary_operator_handler(self, operator, t, handler):
        self._operators[(operator, t)] = handler

    def dispatch_unary_operator(self, operator, t, term):
        try:
            return self._operators[(operator, t)](term)
        except KeyError:
            raise err.LanguageError("Operator '{}' not defined on domain ({})".format(operator, t))

    def dispatch_operator(self, operator, t1, t2, lhs, rhs):
        # assert isinstance(lhs, t1)
        # assert isinstance(rhs, t2)
        op = self._operators.get((operator, t1, t2), None)
        if op is None:
            raise err.LanguageError("Operator '{}' not defined on domain ({}, {})".format(operator, t1, t2))

        return op(lhs, rhs)

    def get(self, first, *args):
        """ Return the language element with given name(s).
        This can be a predicate or function symbol, including constants, or a sort name."""
        res = []
        for what in itertools.chain([first], args):
            try:
                res.append(self._global_index[what])
            except KeyError:
                raise err.UndefinedElement(what)

        return res[0] if not args else res  # Unpack the result if only one element



# Generated from /home/frances/projects/code/tarski/utils/parsers/grammars/fstrips.g4 by ANTLR 4.7.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .parser import fstripsParser
else:
    from parser import fstripsParser

# This class defines a complete listener for a parse tree produced by fstripsParser.
class fstripsListener(ParseTreeListener):

    # Enter a parse tree produced by fstripsParser#pddlDoc.
    def enterPddlDoc(self, ctx:fstripsParser.PddlDocContext):
        pass

    # Exit a parse tree produced by fstripsParser#pddlDoc.
    def exitPddlDoc(self, ctx:fstripsParser.PddlDocContext):
        pass


    # Enter a parse tree produced by fstripsParser#domain.
    def enterDomain(self, ctx:fstripsParser.DomainContext):
        pass

    # Exit a parse tree produced by fstripsParser#domain.
    def exitDomain(self, ctx:fstripsParser.DomainContext):
        pass


    # Enter a parse tree produced by fstripsParser#domainName.
    def enterDomainName(self, ctx:fstripsParser.DomainNameContext):
        pass

    # Exit a parse tree produced by fstripsParser#domainName.
    def exitDomainName(self, ctx:fstripsParser.DomainNameContext):
        pass


    # Enter a parse tree produced by fstripsParser#requireDef.
    def enterRequireDef(self, ctx:fstripsParser.RequireDefContext):
        pass

    # Exit a parse tree produced by fstripsParser#requireDef.
    def exitRequireDef(self, ctx:fstripsParser.RequireDefContext):
        pass


    # Enter a parse tree produced by fstripsParser#declaration_of_types.
    def enterDeclaration_of_types(self, ctx:fstripsParser.Declaration_of_typesContext):
        pass

    # Exit a parse tree produced by fstripsParser#declaration_of_types.
    def exitDeclaration_of_types(self, ctx:fstripsParser.Declaration_of_typesContext):
        pass


    # Enter a parse tree produced by fstripsParser#Integer.
    def enterInteger(self, ctx:fstripsParser.IntegerContext):
        pass

    # Exit a parse tree produced by fstripsParser#Integer.
    def exitInteger(self, ctx:fstripsParser.IntegerContext):
        pass


    # Enter a parse tree produced by fstripsParser#Float.
    def enterFloat(self, ctx:fstripsParser.FloatContext):
        pass

    # Exit a parse tree produced by fstripsParser#Float.
    def exitFloat(self, ctx:fstripsParser.FloatContext):
        pass


    # Enter a parse tree produced by fstripsParser#Number.
    def enterNumber(self, ctx:fstripsParser.NumberContext):
        pass

    # Exit a parse tree produced by fstripsParser#Number.
    def exitNumber(self, ctx:fstripsParser.NumberContext):
        pass


    # Enter a parse tree produced by fstripsParser#NumericBuiltin.
    def enterNumericBuiltin(self, ctx:fstripsParser.NumericBuiltinContext):
        pass

    # Exit a parse tree produced by fstripsParser#NumericBuiltin.
    def exitNumericBuiltin(self, ctx:fstripsParser.NumericBuiltinContext):
        pass


    # Enter a parse tree produced by fstripsParser#ObjectBuiltin.
    def enterObjectBuiltin(self, ctx:fstripsParser.ObjectBuiltinContext):
        pass

    # Exit a parse tree produced by fstripsParser#ObjectBuiltin.
    def exitObjectBuiltin(self, ctx:fstripsParser.ObjectBuiltinContext):
        pass


    # Enter a parse tree produced by fstripsParser#SimpleNameList.
    def enterSimpleNameList(self, ctx:fstripsParser.SimpleNameListContext):
        pass

    # Exit a parse tree produced by fstripsParser#SimpleNameList.
    def exitSimpleNameList(self, ctx:fstripsParser.SimpleNameListContext):
        pass


    # Enter a parse tree produced by fstripsParser#ComplexNameList.
    def enterComplexNameList(self, ctx:fstripsParser.ComplexNameListContext):
        pass

    # Exit a parse tree produced by fstripsParser#ComplexNameList.
    def exitComplexNameList(self, ctx:fstripsParser.ComplexNameListContext):
        pass


    # Enter a parse tree produced by fstripsParser#name_list_with_type.
    def enterName_list_with_type(self, ctx:fstripsParser.Name_list_with_typeContext):
        pass

    # Exit a parse tree produced by fstripsParser#name_list_with_type.
    def exitName_list_with_type(self, ctx:fstripsParser.Name_list_with_typeContext):
        pass


    # Enter a parse tree produced by fstripsParser#UntypedTypenameList.
    def enterUntypedTypenameList(self, ctx:fstripsParser.UntypedTypenameListContext):
        pass

    # Exit a parse tree produced by fstripsParser#UntypedTypenameList.
    def exitUntypedTypenameList(self, ctx:fstripsParser.UntypedTypenameListContext):
        pass


    # Enter a parse tree produced by fstripsParser#TypedTypenameList.
    def enterTypedTypenameList(self, ctx:fstripsParser.TypedTypenameListContext):
        pass

    # Exit a parse tree produced by fstripsParser#TypedTypenameList.
    def exitTypedTypenameList(self, ctx:fstripsParser.TypedTypenameListContext):
        pass


    # Enter a parse tree produced by fstripsParser#UntypedVariableList.
    def enterUntypedVariableList(self, ctx:fstripsParser.UntypedVariableListContext):
        pass

    # Exit a parse tree produced by fstripsParser#UntypedVariableList.
    def exitUntypedVariableList(self, ctx:fstripsParser.UntypedVariableListContext):
        pass


    # Enter a parse tree produced by fstripsParser#TypedVariableList.
    def enterTypedVariableList(self, ctx:fstripsParser.TypedVariableListContext):
        pass

    # Exit a parse tree produced by fstripsParser#TypedVariableList.
    def exitTypedVariableList(self, ctx:fstripsParser.TypedVariableListContext):
        pass


    # Enter a parse tree produced by fstripsParser#variable_list_with_type.
    def enterVariable_list_with_type(self, ctx:fstripsParser.Variable_list_with_typeContext):
        pass

    # Exit a parse tree produced by fstripsParser#variable_list_with_type.
    def exitVariable_list_with_type(self, ctx:fstripsParser.Variable_list_with_typeContext):
        pass


    # Enter a parse tree produced by fstripsParser#EitherTypename.
    def enterEitherTypename(self, ctx:fstripsParser.EitherTypenameContext):
        pass

    # Exit a parse tree produced by fstripsParser#EitherTypename.
    def exitEitherTypename(self, ctx:fstripsParser.EitherTypenameContext):
        pass


    # Enter a parse tree produced by fstripsParser#PrimitiveTypename.
    def enterPrimitiveTypename(self, ctx:fstripsParser.PrimitiveTypenameContext):
        pass

    # Exit a parse tree produced by fstripsParser#PrimitiveTypename.
    def exitPrimitiveTypename(self, ctx:fstripsParser.PrimitiveTypenameContext):
        pass


    # Enter a parse tree produced by fstripsParser#primitive_type.
    def enterPrimitive_type(self, ctx:fstripsParser.Primitive_typeContext):
        pass

    # Exit a parse tree produced by fstripsParser#primitive_type.
    def exitPrimitive_type(self, ctx:fstripsParser.Primitive_typeContext):
        pass


    # Enter a parse tree produced by fstripsParser#function_definition_block.
    def enterFunction_definition_block(self, ctx:fstripsParser.Function_definition_blockContext):
        pass

    # Exit a parse tree produced by fstripsParser#function_definition_block.
    def exitFunction_definition_block(self, ctx:fstripsParser.Function_definition_blockContext):
        pass


    # Enter a parse tree produced by fstripsParser#single_function_definition.
    def enterSingle_function_definition(self, ctx:fstripsParser.Single_function_definitionContext):
        pass

    # Exit a parse tree produced by fstripsParser#single_function_definition.
    def exitSingle_function_definition(self, ctx:fstripsParser.Single_function_definitionContext):
        pass


    # Enter a parse tree produced by fstripsParser#typed_function_definition.
    def enterTyped_function_definition(self, ctx:fstripsParser.Typed_function_definitionContext):
        pass

    # Exit a parse tree produced by fstripsParser#typed_function_definition.
    def exitTyped_function_definition(self, ctx:fstripsParser.Typed_function_definitionContext):
        pass


    # Enter a parse tree produced by fstripsParser#untyped_function_definition.
    def enterUntyped_function_definition(self, ctx:fstripsParser.Untyped_function_definitionContext):
        pass

    # Exit a parse tree produced by fstripsParser#untyped_function_definition.
    def exitUntyped_function_definition(self, ctx:fstripsParser.Untyped_function_definitionContext):
        pass


    # Enter a parse tree produced by fstripsParser#logical_symbol_name.
    def enterLogical_symbol_name(self, ctx:fstripsParser.Logical_symbol_nameContext):
        pass

    # Exit a parse tree produced by fstripsParser#logical_symbol_name.
    def exitLogical_symbol_name(self, ctx:fstripsParser.Logical_symbol_nameContext):
        pass


    # Enter a parse tree produced by fstripsParser#constant_declaration.
    def enterConstant_declaration(self, ctx:fstripsParser.Constant_declarationContext):
        pass

    # Exit a parse tree produced by fstripsParser#constant_declaration.
    def exitConstant_declaration(self, ctx:fstripsParser.Constant_declarationContext):
        pass


    # Enter a parse tree produced by fstripsParser#predicate_definition_block.
    def enterPredicate_definition_block(self, ctx:fstripsParser.Predicate_definition_blockContext):
        pass

    # Exit a parse tree produced by fstripsParser#predicate_definition_block.
    def exitPredicate_definition_block(self, ctx:fstripsParser.Predicate_definition_blockContext):
        pass


    # Enter a parse tree produced by fstripsParser#single_predicate_definition.
    def enterSingle_predicate_definition(self, ctx:fstripsParser.Single_predicate_definitionContext):
        pass

    # Exit a parse tree produced by fstripsParser#single_predicate_definition.
    def exitSingle_predicate_definition(self, ctx:fstripsParser.Single_predicate_definitionContext):
        pass


    # Enter a parse tree produced by fstripsParser#predicate.
    def enterPredicate(self, ctx:fstripsParser.PredicateContext):
        pass

    # Exit a parse tree produced by fstripsParser#predicate.
    def exitPredicate(self, ctx:fstripsParser.PredicateContext):
        pass


    # Enter a parse tree produced by fstripsParser#function_name.
    def enterFunction_name(self, ctx:fstripsParser.Function_nameContext):
        pass

    # Exit a parse tree produced by fstripsParser#function_name.
    def exitFunction_name(self, ctx:fstripsParser.Function_nameContext):
        pass


    # Enter a parse tree produced by fstripsParser#structureDef.
    def enterStructureDef(self, ctx:fstripsParser.StructureDefContext):
        pass

    # Exit a parse tree produced by fstripsParser#structureDef.
    def exitStructureDef(self, ctx:fstripsParser.StructureDefContext):
        pass


    # Enter a parse tree produced by fstripsParser#actionDef.
    def enterActionDef(self, ctx:fstripsParser.ActionDefContext):
        pass

    # Exit a parse tree produced by fstripsParser#actionDef.
    def exitActionDef(self, ctx:fstripsParser.ActionDefContext):
        pass


    # Enter a parse tree produced by fstripsParser#constraintDef.
    def enterConstraintDef(self, ctx:fstripsParser.ConstraintDefContext):
        pass

    # Exit a parse tree produced by fstripsParser#constraintDef.
    def exitConstraintDef(self, ctx:fstripsParser.ConstraintDefContext):
        pass


    # Enter a parse tree produced by fstripsParser#eventDef.
    def enterEventDef(self, ctx:fstripsParser.EventDefContext):
        pass

    # Exit a parse tree produced by fstripsParser#eventDef.
    def exitEventDef(self, ctx:fstripsParser.EventDefContext):
        pass


    # Enter a parse tree produced by fstripsParser#actionName.
    def enterActionName(self, ctx:fstripsParser.ActionNameContext):
        pass

    # Exit a parse tree produced by fstripsParser#actionName.
    def exitActionName(self, ctx:fstripsParser.ActionNameContext):
        pass


    # Enter a parse tree produced by fstripsParser#constraintSymbol.
    def enterConstraintSymbol(self, ctx:fstripsParser.ConstraintSymbolContext):
        pass

    # Exit a parse tree produced by fstripsParser#constraintSymbol.
    def exitConstraintSymbol(self, ctx:fstripsParser.ConstraintSymbolContext):
        pass


    # Enter a parse tree produced by fstripsParser#eventSymbol.
    def enterEventSymbol(self, ctx:fstripsParser.EventSymbolContext):
        pass

    # Exit a parse tree produced by fstripsParser#eventSymbol.
    def exitEventSymbol(self, ctx:fstripsParser.EventSymbolContext):
        pass


    # Enter a parse tree produced by fstripsParser#actionDefBody.
    def enterActionDefBody(self, ctx:fstripsParser.ActionDefBodyContext):
        pass

    # Exit a parse tree produced by fstripsParser#actionDefBody.
    def exitActionDefBody(self, ctx:fstripsParser.ActionDefBodyContext):
        pass


    # Enter a parse tree produced by fstripsParser#TrivialPrecondition.
    def enterTrivialPrecondition(self, ctx:fstripsParser.TrivialPreconditionContext):
        pass

    # Exit a parse tree produced by fstripsParser#TrivialPrecondition.
    def exitTrivialPrecondition(self, ctx:fstripsParser.TrivialPreconditionContext):
        pass


    # Enter a parse tree produced by fstripsParser#RegularPrecondition.
    def enterRegularPrecondition(self, ctx:fstripsParser.RegularPreconditionContext):
        pass

    # Exit a parse tree produced by fstripsParser#RegularPrecondition.
    def exitRegularPrecondition(self, ctx:fstripsParser.RegularPreconditionContext):
        pass


    # Enter a parse tree produced by fstripsParser#TermGoalDesc.
    def enterTermGoalDesc(self, ctx:fstripsParser.TermGoalDescContext):
        pass

    # Exit a parse tree produced by fstripsParser#TermGoalDesc.
    def exitTermGoalDesc(self, ctx:fstripsParser.TermGoalDescContext):
        pass


    # Enter a parse tree produced by fstripsParser#AndGoalDesc.
    def enterAndGoalDesc(self, ctx:fstripsParser.AndGoalDescContext):
        pass

    # Exit a parse tree produced by fstripsParser#AndGoalDesc.
    def exitAndGoalDesc(self, ctx:fstripsParser.AndGoalDescContext):
        pass


    # Enter a parse tree produced by fstripsParser#OrGoalDesc.
    def enterOrGoalDesc(self, ctx:fstripsParser.OrGoalDescContext):
        pass

    # Exit a parse tree produced by fstripsParser#OrGoalDesc.
    def exitOrGoalDesc(self, ctx:fstripsParser.OrGoalDescContext):
        pass


    # Enter a parse tree produced by fstripsParser#NotGoalDesc.
    def enterNotGoalDesc(self, ctx:fstripsParser.NotGoalDescContext):
        pass

    # Exit a parse tree produced by fstripsParser#NotGoalDesc.
    def exitNotGoalDesc(self, ctx:fstripsParser.NotGoalDescContext):
        pass


    # Enter a parse tree produced by fstripsParser#ImplyGoalDesc.
    def enterImplyGoalDesc(self, ctx:fstripsParser.ImplyGoalDescContext):
        pass

    # Exit a parse tree produced by fstripsParser#ImplyGoalDesc.
    def exitImplyGoalDesc(self, ctx:fstripsParser.ImplyGoalDescContext):
        pass


    # Enter a parse tree produced by fstripsParser#ExistentialGoalDesc.
    def enterExistentialGoalDesc(self, ctx:fstripsParser.ExistentialGoalDescContext):
        pass

    # Exit a parse tree produced by fstripsParser#ExistentialGoalDesc.
    def exitExistentialGoalDesc(self, ctx:fstripsParser.ExistentialGoalDescContext):
        pass


    # Enter a parse tree produced by fstripsParser#UniversalGoalDesc.
    def enterUniversalGoalDesc(self, ctx:fstripsParser.UniversalGoalDescContext):
        pass

    # Exit a parse tree produced by fstripsParser#UniversalGoalDesc.
    def exitUniversalGoalDesc(self, ctx:fstripsParser.UniversalGoalDescContext):
        pass


    # Enter a parse tree produced by fstripsParser#BuiltinBinaryAtom.
    def enterBuiltinBinaryAtom(self, ctx:fstripsParser.BuiltinBinaryAtomContext):
        pass

    # Exit a parse tree produced by fstripsParser#BuiltinBinaryAtom.
    def exitBuiltinBinaryAtom(self, ctx:fstripsParser.BuiltinBinaryAtomContext):
        pass


    # Enter a parse tree produced by fstripsParser#atomicTermFormula.
    def enterAtomicTermFormula(self, ctx:fstripsParser.AtomicTermFormulaContext):
        pass

    # Exit a parse tree produced by fstripsParser#atomicTermFormula.
    def exitAtomicTermFormula(self, ctx:fstripsParser.AtomicTermFormulaContext):
        pass


    # Enter a parse tree produced by fstripsParser#TermObject.
    def enterTermObject(self, ctx:fstripsParser.TermObjectContext):
        pass

    # Exit a parse tree produced by fstripsParser#TermObject.
    def exitTermObject(self, ctx:fstripsParser.TermObjectContext):
        pass


    # Enter a parse tree produced by fstripsParser#TermNumber.
    def enterTermNumber(self, ctx:fstripsParser.TermNumberContext):
        pass

    # Exit a parse tree produced by fstripsParser#TermNumber.
    def exitTermNumber(self, ctx:fstripsParser.TermNumberContext):
        pass


    # Enter a parse tree produced by fstripsParser#TermVariable.
    def enterTermVariable(self, ctx:fstripsParser.TermVariableContext):
        pass

    # Exit a parse tree produced by fstripsParser#TermVariable.
    def exitTermVariable(self, ctx:fstripsParser.TermVariableContext):
        pass


    # Enter a parse tree produced by fstripsParser#TermTimeStep.
    def enterTermTimeStep(self, ctx:fstripsParser.TermTimeStepContext):
        pass

    # Exit a parse tree produced by fstripsParser#TermTimeStep.
    def exitTermTimeStep(self, ctx:fstripsParser.TermTimeStepContext):
        pass


    # Enter a parse tree produced by fstripsParser#TermFunction.
    def enterTermFunction(self, ctx:fstripsParser.TermFunctionContext):
        pass

    # Exit a parse tree produced by fstripsParser#TermFunction.
    def exitTermFunction(self, ctx:fstripsParser.TermFunctionContext):
        pass


    # Enter a parse tree produced by fstripsParser#GenericFunctionTerm.
    def enterGenericFunctionTerm(self, ctx:fstripsParser.GenericFunctionTermContext):
        pass

    # Exit a parse tree produced by fstripsParser#GenericFunctionTerm.
    def exitGenericFunctionTerm(self, ctx:fstripsParser.GenericFunctionTermContext):
        pass


    # Enter a parse tree produced by fstripsParser#BinaryArithmeticFunctionTerm.
    def enterBinaryArithmeticFunctionTerm(self, ctx:fstripsParser.BinaryArithmeticFunctionTermContext):
        pass

    # Exit a parse tree produced by fstripsParser#BinaryArithmeticFunctionTerm.
    def exitBinaryArithmeticFunctionTerm(self, ctx:fstripsParser.BinaryArithmeticFunctionTermContext):
        pass


    # Enter a parse tree produced by fstripsParser#UnaryArithmeticFunctionTerm.
    def enterUnaryArithmeticFunctionTerm(self, ctx:fstripsParser.UnaryArithmeticFunctionTermContext):
        pass

    # Exit a parse tree produced by fstripsParser#UnaryArithmeticFunctionTerm.
    def exitUnaryArithmeticFunctionTerm(self, ctx:fstripsParser.UnaryArithmeticFunctionTermContext):
        pass


    # Enter a parse tree produced by fstripsParser#derivedDef.
    def enterDerivedDef(self, ctx:fstripsParser.DerivedDefContext):
        pass

    # Exit a parse tree produced by fstripsParser#derivedDef.
    def exitDerivedDef(self, ctx:fstripsParser.DerivedDefContext):
        pass


    # Enter a parse tree produced by fstripsParser#ConjunctiveEffectFormula.
    def enterConjunctiveEffectFormula(self, ctx:fstripsParser.ConjunctiveEffectFormulaContext):
        pass

    # Exit a parse tree produced by fstripsParser#ConjunctiveEffectFormula.
    def exitConjunctiveEffectFormula(self, ctx:fstripsParser.ConjunctiveEffectFormulaContext):
        pass


    # Enter a parse tree produced by fstripsParser#SingleEffect.
    def enterSingleEffect(self, ctx:fstripsParser.SingleEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#SingleEffect.
    def exitSingleEffect(self, ctx:fstripsParser.SingleEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#UniversallyQuantifiedEffect.
    def enterUniversallyQuantifiedEffect(self, ctx:fstripsParser.UniversallyQuantifiedEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#UniversallyQuantifiedEffect.
    def exitUniversallyQuantifiedEffect(self, ctx:fstripsParser.UniversallyQuantifiedEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#SingleConditionalEffect.
    def enterSingleConditionalEffect(self, ctx:fstripsParser.SingleConditionalEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#SingleConditionalEffect.
    def exitSingleConditionalEffect(self, ctx:fstripsParser.SingleConditionalEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#MultipleConditionalEffect.
    def enterMultipleConditionalEffect(self, ctx:fstripsParser.MultipleConditionalEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#MultipleConditionalEffect.
    def exitMultipleConditionalEffect(self, ctx:fstripsParser.MultipleConditionalEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#AtomicEffect.
    def enterAtomicEffect(self, ctx:fstripsParser.AtomicEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#AtomicEffect.
    def exitAtomicEffect(self, ctx:fstripsParser.AtomicEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#AssignConstant.
    def enterAssignConstant(self, ctx:fstripsParser.AssignConstantContext):
        pass

    # Exit a parse tree produced by fstripsParser#AssignConstant.
    def exitAssignConstant(self, ctx:fstripsParser.AssignConstantContext):
        pass


    # Enter a parse tree produced by fstripsParser#AssignEffect.
    def enterAssignEffect(self, ctx:fstripsParser.AssignEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#AssignEffect.
    def exitAssignEffect(self, ctx:fstripsParser.AssignEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#DeleteAtomEffect.
    def enterDeleteAtomEffect(self, ctx:fstripsParser.DeleteAtomEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#DeleteAtomEffect.
    def exitDeleteAtomEffect(self, ctx:fstripsParser.DeleteAtomEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#AddAtomEffect.
    def enterAddAtomEffect(self, ctx:fstripsParser.AddAtomEffectContext):
        pass

    # Exit a parse tree produced by fstripsParser#AddAtomEffect.
    def exitAddAtomEffect(self, ctx:fstripsParser.AddAtomEffectContext):
        pass


    # Enter a parse tree produced by fstripsParser#builtin_binary_function.
    def enterBuiltin_binary_function(self, ctx:fstripsParser.Builtin_binary_functionContext):
        pass

    # Exit a parse tree produced by fstripsParser#builtin_binary_function.
    def exitBuiltin_binary_function(self, ctx:fstripsParser.Builtin_binary_functionContext):
        pass


    # Enter a parse tree produced by fstripsParser#builtin_unary_function.
    def enterBuiltin_unary_function(self, ctx:fstripsParser.Builtin_unary_functionContext):
        pass

    # Exit a parse tree produced by fstripsParser#builtin_unary_function.
    def exitBuiltin_unary_function(self, ctx:fstripsParser.Builtin_unary_functionContext):
        pass


    # Enter a parse tree produced by fstripsParser#builtin_binary_predicate.
    def enterBuiltin_binary_predicate(self, ctx:fstripsParser.Builtin_binary_predicateContext):
        pass

    # Exit a parse tree produced by fstripsParser#builtin_binary_predicate.
    def exitBuiltin_binary_predicate(self, ctx:fstripsParser.Builtin_binary_predicateContext):
        pass


    # Enter a parse tree produced by fstripsParser#assignOp.
    def enterAssignOp(self, ctx:fstripsParser.AssignOpContext):
        pass

    # Exit a parse tree produced by fstripsParser#assignOp.
    def exitAssignOp(self, ctx:fstripsParser.AssignOpContext):
        pass


    # Enter a parse tree produced by fstripsParser#problem.
    def enterProblem(self, ctx:fstripsParser.ProblemContext):
        pass

    # Exit a parse tree produced by fstripsParser#problem.
    def exitProblem(self, ctx:fstripsParser.ProblemContext):
        pass


    # Enter a parse tree produced by fstripsParser#problemMeta.
    def enterProblemMeta(self, ctx:fstripsParser.ProblemMetaContext):
        pass

    # Exit a parse tree produced by fstripsParser#problemMeta.
    def exitProblemMeta(self, ctx:fstripsParser.ProblemMetaContext):
        pass


    # Enter a parse tree produced by fstripsParser#problemDecl.
    def enterProblemDecl(self, ctx:fstripsParser.ProblemDeclContext):
        pass

    # Exit a parse tree produced by fstripsParser#problemDecl.
    def exitProblemDecl(self, ctx:fstripsParser.ProblemDeclContext):
        pass


    # Enter a parse tree produced by fstripsParser#problemDomain.
    def enterProblemDomain(self, ctx:fstripsParser.ProblemDomainContext):
        pass

    # Exit a parse tree produced by fstripsParser#problemDomain.
    def exitProblemDomain(self, ctx:fstripsParser.ProblemDomainContext):
        pass


    # Enter a parse tree produced by fstripsParser#object_declaration.
    def enterObject_declaration(self, ctx:fstripsParser.Object_declarationContext):
        pass

    # Exit a parse tree produced by fstripsParser#object_declaration.
    def exitObject_declaration(self, ctx:fstripsParser.Object_declarationContext):
        pass


    # Enter a parse tree produced by fstripsParser#boundsDecl.
    def enterBoundsDecl(self, ctx:fstripsParser.BoundsDeclContext):
        pass

    # Exit a parse tree produced by fstripsParser#boundsDecl.
    def exitBoundsDecl(self, ctx:fstripsParser.BoundsDeclContext):
        pass


    # Enter a parse tree produced by fstripsParser#typeBoundsDefinition.
    def enterTypeBoundsDefinition(self, ctx:fstripsParser.TypeBoundsDefinitionContext):
        pass

    # Exit a parse tree produced by fstripsParser#typeBoundsDefinition.
    def exitTypeBoundsDefinition(self, ctx:fstripsParser.TypeBoundsDefinitionContext):
        pass


    # Enter a parse tree produced by fstripsParser#init.
    def enterInit(self, ctx:fstripsParser.InitContext):
        pass

    # Exit a parse tree produced by fstripsParser#init.
    def exitInit(self, ctx:fstripsParser.InitContext):
        pass


    # Enter a parse tree produced by fstripsParser#InitPositiveLiteral.
    def enterInitPositiveLiteral(self, ctx:fstripsParser.InitPositiveLiteralContext):
        pass

    # Exit a parse tree produced by fstripsParser#InitPositiveLiteral.
    def exitInitPositiveLiteral(self, ctx:fstripsParser.InitPositiveLiteralContext):
        pass


    # Enter a parse tree produced by fstripsParser#InitNegativeLiteral.
    def enterInitNegativeLiteral(self, ctx:fstripsParser.InitNegativeLiteralContext):
        pass

    # Exit a parse tree produced by fstripsParser#InitNegativeLiteral.
    def exitInitNegativeLiteral(self, ctx:fstripsParser.InitNegativeLiteralContext):
        pass


    # Enter a parse tree produced by fstripsParser#InitFunctionAssignment.
    def enterInitFunctionAssignment(self, ctx:fstripsParser.InitFunctionAssignmentContext):
        pass

    # Exit a parse tree produced by fstripsParser#InitFunctionAssignment.
    def exitInitFunctionAssignment(self, ctx:fstripsParser.InitFunctionAssignmentContext):
        pass


    # Enter a parse tree produced by fstripsParser#flat_term.
    def enterFlat_term(self, ctx:fstripsParser.Flat_termContext):
        pass

    # Exit a parse tree produced by fstripsParser#flat_term.
    def exitFlat_term(self, ctx:fstripsParser.Flat_termContext):
        pass


    # Enter a parse tree produced by fstripsParser#flat_atom.
    def enterFlat_atom(self, ctx:fstripsParser.Flat_atomContext):
        pass

    # Exit a parse tree produced by fstripsParser#flat_atom.
    def exitFlat_atom(self, ctx:fstripsParser.Flat_atomContext):
        pass


    # Enter a parse tree produced by fstripsParser#symbolic_constant.
    def enterSymbolic_constant(self, ctx:fstripsParser.Symbolic_constantContext):
        pass

    # Exit a parse tree produced by fstripsParser#symbolic_constant.
    def exitSymbolic_constant(self, ctx:fstripsParser.Symbolic_constantContext):
        pass


    # Enter a parse tree produced by fstripsParser#numeric_constant.
    def enterNumeric_constant(self, ctx:fstripsParser.Numeric_constantContext):
        pass

    # Exit a parse tree produced by fstripsParser#numeric_constant.
    def exitNumeric_constant(self, ctx:fstripsParser.Numeric_constantContext):
        pass


    # Enter a parse tree produced by fstripsParser#goal.
    def enterGoal(self, ctx:fstripsParser.GoalContext):
        pass

    # Exit a parse tree produced by fstripsParser#goal.
    def exitGoal(self, ctx:fstripsParser.GoalContext):
        pass


    # Enter a parse tree produced by fstripsParser#probConstraints.
    def enterProbConstraints(self, ctx:fstripsParser.ProbConstraintsContext):
        pass

    # Exit a parse tree produced by fstripsParser#probConstraints.
    def exitProbConstraints(self, ctx:fstripsParser.ProbConstraintsContext):
        pass


    # Enter a parse tree produced by fstripsParser#ConjunctionOfConstraints.
    def enterConjunctionOfConstraints(self, ctx:fstripsParser.ConjunctionOfConstraintsContext):
        pass

    # Exit a parse tree produced by fstripsParser#ConjunctionOfConstraints.
    def exitConjunctionOfConstraints(self, ctx:fstripsParser.ConjunctionOfConstraintsContext):
        pass


    # Enter a parse tree produced by fstripsParser#UniversallyQuantifiedConstraint.
    def enterUniversallyQuantifiedConstraint(self, ctx:fstripsParser.UniversallyQuantifiedConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#UniversallyQuantifiedConstraint.
    def exitUniversallyQuantifiedConstraint(self, ctx:fstripsParser.UniversallyQuantifiedConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#PreferenceConstraint.
    def enterPreferenceConstraint(self, ctx:fstripsParser.PreferenceConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#PreferenceConstraint.
    def exitPreferenceConstraint(self, ctx:fstripsParser.PreferenceConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#PlainConstraintList.
    def enterPlainConstraintList(self, ctx:fstripsParser.PlainConstraintListContext):
        pass

    # Exit a parse tree produced by fstripsParser#PlainConstraintList.
    def exitPlainConstraintList(self, ctx:fstripsParser.PlainConstraintListContext):
        pass


    # Enter a parse tree produced by fstripsParser#ProblemMetric.
    def enterProblemMetric(self, ctx:fstripsParser.ProblemMetricContext):
        pass

    # Exit a parse tree produced by fstripsParser#ProblemMetric.
    def exitProblemMetric(self, ctx:fstripsParser.ProblemMetricContext):
        pass


    # Enter a parse tree produced by fstripsParser#optimization.
    def enterOptimization(self, ctx:fstripsParser.OptimizationContext):
        pass

    # Exit a parse tree produced by fstripsParser#optimization.
    def exitOptimization(self, ctx:fstripsParser.OptimizationContext):
        pass


    # Enter a parse tree produced by fstripsParser#FunctionalExprMetric.
    def enterFunctionalExprMetric(self, ctx:fstripsParser.FunctionalExprMetricContext):
        pass

    # Exit a parse tree produced by fstripsParser#FunctionalExprMetric.
    def exitFunctionalExprMetric(self, ctx:fstripsParser.FunctionalExprMetricContext):
        pass


    # Enter a parse tree produced by fstripsParser#CompositeMetric.
    def enterCompositeMetric(self, ctx:fstripsParser.CompositeMetricContext):
        pass

    # Exit a parse tree produced by fstripsParser#CompositeMetric.
    def exitCompositeMetric(self, ctx:fstripsParser.CompositeMetricContext):
        pass


    # Enter a parse tree produced by fstripsParser#TotalTimeMetric.
    def enterTotalTimeMetric(self, ctx:fstripsParser.TotalTimeMetricContext):
        pass

    # Exit a parse tree produced by fstripsParser#TotalTimeMetric.
    def exitTotalTimeMetric(self, ctx:fstripsParser.TotalTimeMetricContext):
        pass


    # Enter a parse tree produced by fstripsParser#IsViolatedMetric.
    def enterIsViolatedMetric(self, ctx:fstripsParser.IsViolatedMetricContext):
        pass

    # Exit a parse tree produced by fstripsParser#IsViolatedMetric.
    def exitIsViolatedMetric(self, ctx:fstripsParser.IsViolatedMetricContext):
        pass


    # Enter a parse tree produced by fstripsParser#terminalCost.
    def enterTerminalCost(self, ctx:fstripsParser.TerminalCostContext):
        pass

    # Exit a parse tree produced by fstripsParser#terminalCost.
    def exitTerminalCost(self, ctx:fstripsParser.TerminalCostContext):
        pass


    # Enter a parse tree produced by fstripsParser#stageCost.
    def enterStageCost(self, ctx:fstripsParser.StageCostContext):
        pass

    # Exit a parse tree produced by fstripsParser#stageCost.
    def exitStageCost(self, ctx:fstripsParser.StageCostContext):
        pass


    # Enter a parse tree produced by fstripsParser#ConjunctiveConstraint.
    def enterConjunctiveConstraint(self, ctx:fstripsParser.ConjunctiveConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#ConjunctiveConstraint.
    def exitConjunctiveConstraint(self, ctx:fstripsParser.ConjunctiveConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#ForallConstraint.
    def enterForallConstraint(self, ctx:fstripsParser.ForallConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#ForallConstraint.
    def exitForallConstraint(self, ctx:fstripsParser.ForallConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#AtEndConstraint.
    def enterAtEndConstraint(self, ctx:fstripsParser.AtEndConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#AtEndConstraint.
    def exitAtEndConstraint(self, ctx:fstripsParser.AtEndConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#AlwaysConstraint.
    def enterAlwaysConstraint(self, ctx:fstripsParser.AlwaysConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#AlwaysConstraint.
    def exitAlwaysConstraint(self, ctx:fstripsParser.AlwaysConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#SometimeConstraint.
    def enterSometimeConstraint(self, ctx:fstripsParser.SometimeConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#SometimeConstraint.
    def exitSometimeConstraint(self, ctx:fstripsParser.SometimeConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#WithinConstraint.
    def enterWithinConstraint(self, ctx:fstripsParser.WithinConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#WithinConstraint.
    def exitWithinConstraint(self, ctx:fstripsParser.WithinConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#AtMostOnceConstraint.
    def enterAtMostOnceConstraint(self, ctx:fstripsParser.AtMostOnceConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#AtMostOnceConstraint.
    def exitAtMostOnceConstraint(self, ctx:fstripsParser.AtMostOnceConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#SometimeAfterConstraint.
    def enterSometimeAfterConstraint(self, ctx:fstripsParser.SometimeAfterConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#SometimeAfterConstraint.
    def exitSometimeAfterConstraint(self, ctx:fstripsParser.SometimeAfterConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#SometimeBeforeConstraint.
    def enterSometimeBeforeConstraint(self, ctx:fstripsParser.SometimeBeforeConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#SometimeBeforeConstraint.
    def exitSometimeBeforeConstraint(self, ctx:fstripsParser.SometimeBeforeConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#AlwaysWithinConstraint.
    def enterAlwaysWithinConstraint(self, ctx:fstripsParser.AlwaysWithinConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#AlwaysWithinConstraint.
    def exitAlwaysWithinConstraint(self, ctx:fstripsParser.AlwaysWithinConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#HoldDuringConstraint.
    def enterHoldDuringConstraint(self, ctx:fstripsParser.HoldDuringConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#HoldDuringConstraint.
    def exitHoldDuringConstraint(self, ctx:fstripsParser.HoldDuringConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#HoldAfterConstraint.
    def enterHoldAfterConstraint(self, ctx:fstripsParser.HoldAfterConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#HoldAfterConstraint.
    def exitHoldAfterConstraint(self, ctx:fstripsParser.HoldAfterConstraintContext):
        pass


    # Enter a parse tree produced by fstripsParser#AlternativeAlwaysConstraint.
    def enterAlternativeAlwaysConstraint(self, ctx:fstripsParser.AlternativeAlwaysConstraintContext):
        pass

    # Exit a parse tree produced by fstripsParser#AlternativeAlwaysConstraint.
    def exitAlternativeAlwaysConstraint(self, ctx:fstripsParser.AlternativeAlwaysConstraintContext):
        pass






from ..errors import TarskiError


class TarskiDLError(TarskiError):
    """ Common ancestor class to all of Tarski's Description Language exceptions """


class SyntacticDLError(TarskiDLError):
    def __init__(self, msg=None):
        msg = msg or 'Unexplained Tarski DL syntactic error'
        super().__init__(msg)

#
# class SemanticDLError(TarskiDLError):
#     def __init__(self, msg=None):
#         msg = msg or 'Unexplained Tarski DL semantic error'
#         super().__init__(msg)


class ArityDLMismatch(SyntacticDLError):
    def __init__(self, msg=None):
        msg = msg or 'Mismatch between language element and DL element'
        super().__init__(msg)




from ...syntax import Interval
from ... import theories
from ...theories import Theory


def pddl_to_tarski_type(typename):
    """ Translate a few PDDL types into their corresponding Tarski names
        (e.g. the FSTRIPS type "int" corresponds to the Tarski type "Integer").
    """
    translations = {"int": "Integer"}
    return translations.get(typename, typename)


def tarski_to_pddl_type(typename):
    """ Translate a few Tarski types into their corresponding FSTRIPS names
    """
    translations = {"Integer": "int", "Natural": "int", "Real": "number"}
    return translations.get(typename.name, typename.name)


def parse_number(number, lang):
    # Because of the grammar, we know that number will be either an int or a float
    try:
        value = int(number)
        return lang.constant(lang.Integer.cast(value), lang.Integer)
    except ValueError:
        value = float(number)
        return lang.constant(lang.Real.cast(value), lang.Real)


def process_requirements(requirements, lang):
    if ':action-costs' in requirements:
        theories.load_theory(lang, Theory.ARITHMETIC)
        create_number_type(lang)
    elif ':numeric-fluents' in requirements:
        theories.load_theory(lang, Theory.ARITHMETIC)


def get_requirements_string(problem):
    """ Get a list with all PDDL requirement strings based on the given problem """
    # TODO To be completed
    requirements = []
    for t in problem.language.theories:
        # The Arithmetic theory incorporates arithmetic functions (which are PDDL numeric operators)
        if t == Theory.ARITHMETIC:
            requirements.append(":numeric-fluents")

    # Functional STRIPS with action costs, keeping in line with IPC convention
    if problem.metric is not None and problem.language.has_function('total-cost'):
        requirements.append(":action-costs")

    return requirements


def create_number_type(lang):
    """ Creates a sort corresponding to the PDDL "number" """
    # For the moment being, we assume that PDDL "number"s are unbound integers
    parent = lang.get_sort("Integer")
    lower = parent.lower_bound
    upper = parent.upper_bound
    lang.interval("number", parent, lower, upper)


def create_sort(lang, typename, basename):
    """ Create a Tarski sort from a PDDL type, performing a few checks and translations to ensure consistency """
    if typename == 'object':
        return
    
    typename = pddl_to_tarski_type(typename)
    basename = pddl_to_tarski_type(basename)

    if basename in ("Natural", "Integer", "Real"):
        # The new type will necesarily be an interval subset of it parent
        parent = lang.get_sort(basename)
        assert isinstance(parent, Interval)
        # TODO This is a small hack: since we don't yet know the lower and upper bound (because they are declared in
        # the instance file, not domain file), we put the widest possible range allowed by the parent. If we do not
        # do that, then the processing of the instance file might raise some error, because when the initial state is
        # parsed, we check that each assignment is within bounds - but that again might happen before knowing the real
        # bounds. A better solution would be to parse the instance file one first pass looking only for the bounds.
        lower = parent.lower_bound
        upper = parent.upper_bound
        lang.interval(typename, parent, lower, upper)
    else:
        lang.sort(typename, lang.get_sort(basename))




from collections import defaultdict

from . import errors as err
from .syntax import Function, Constant, CompoundTerm, symref
from .syntax.predicate import Predicate


def _check_assignment(fun, point, value=None):
    assert isinstance(point, tuple)

    elements = point + (value,) if value is not None else point
    processed = []

    typ = fun.sort

    if len(typ) != len(elements):
        raise err.ArityMismatch(fun, elements)

    language = fun.language
    for element, expected_type in zip(elements, typ):

        if not isinstance(element, Constant):
            # Assume a literal value has been passed instead of its corresponding constant
            element = Constant(expected_type.cast(element), expected_type)
            # raise err.IncorrectExtensionDefinition(fun, point, value)

        if element.language != language:
            raise err.LanguageMismatch(element, element.language, language)

        if not language.is_subtype(element.sort, expected_type):
            raise err.SortMismatch(element, element.sort, expected_type)

        processed.append(element)

    if value is None:
        return tuple(processed)

    assert len(processed) > 0
    return tuple(processed[:-1]), processed[-1]


class Model:
    """ A First Order Language Model """

    def __init__(self, language):
        self.evaluator = None
        self.language = language
        self.function_extensions = dict()
        self.predicate_extensions = defaultdict(set)

    def setx(self, t: CompoundTerm, value: Constant):
        if not isinstance(t.symbol, Function):
            raise err.SemanticError("Model.set() can only set the value of functions")
        if t.symbol.builtin:
            raise err.SemanticError("Model.set() cannot redefine builtin symbols like '{}'".format(str(t.symbol)))
        for st in t.subterms:
            if not isinstance(st, Constant):
                raise err.SemanticError("Model.set(): subterms of '{}' need to be constants".format(str(t)))
        point, value = _check_assignment(t.symbol, tuple(t.subterms), value)
        if t.symbol.signature not in self.function_extensions:
            definition = self.function_extensions[t.symbol.signature] = ExtensionalFunctionDefinition()
        else:
            definition = self.function_extensions[t.symbol.signature]
            if not isinstance(definition, ExtensionalFunctionDefinition):
                raise err.SemanticError("Cannot define extension of intensional definition")

        definition.set(point, value)

    def set(self, fun, *args):
        """ Set the value of fucntion 'fun' at point 'point' to be equal to 'value'
            'point' needs to be a tuple of constants, and value a single constant.
        """
        if not isinstance(fun, Function):
            raise err.SemanticError("Model.set() can only set the value of functions")
        point, value = args[:-1], args[-1]
        point, value = _check_assignment(fun, point, value)
        if fun.signature not in self.function_extensions:
            definition = self.function_extensions[fun.signature] = ExtensionalFunctionDefinition()
        else:
            definition = self.function_extensions[fun.signature]
            if not isinstance(definition, ExtensionalFunctionDefinition):
                raise err.SemanticError("Cannot define extension of intensional definition")

        definition.set(point, value)

    def add(self, predicate: Predicate, *args):
        if not isinstance(predicate, Predicate):
            raise err.SemanticError("Model.add() can only set the value of predicates")
        point = _check_assignment(predicate, args)
        self.predicate_extensions[predicate.signature].add(wrap_tuple(point))

    def remove(self, predicate: Predicate, *args):
        self.predicate_extensions[predicate.signature].remove(wrap_tuple(args))

    def value(self, fun: Function, point):
        """ Return the value of the given function on the given point in the current model """
        definition = self.function_extensions[fun.signature]
        return definition.get(point)

    def holds(self, predicate: Predicate, point):
        """ Return true iff the given predicate is true on the given point in the current model """
        # return tuple(c.symbol for c in point) in self.predicate_extensions[predicate.signature]
        return wrap_tuple(point) in self.predicate_extensions[predicate.signature]

    def list_all_extensions(self):
        """ Return a mapping between predicate and function signatures and a list of all their respective extensions.
        This list *unwraps* the TermReference's used internally in this class back into plain Tarski terms, so that
        you can rely on the returned extensions being made up of Constant's, Variables, etc., not TermReference's
        """
        exts = {k: [unwrap_tuple(tup) for tup in ext] for k, ext in self.predicate_extensions.items()}
        exts.update((k, [unwrap_tuple(point) + (value, ) for point, value in ext.data.items()])
                    for k, ext in self.function_extensions.items())
        return exts

    def as_atoms(self):
        """ Return a representation of the model as a list of atoms that are true. For functional symbols f, return
        tuples of the form (f(o1, ..., on), value)
         """
        atoms = []
        for k, ext in self.predicate_extensions.items():
            pred = self.language.get_predicate(k[0])
            # create one atom per tuple in the predicate's extension:
            atoms += [pred(*unwrap_tuple(tup)) for tup in ext]

        for k, ext in self.function_extensions.items():
            fun = self.language.get_function(k[0])
            for point, value in ext.data.items():
                atoms.append((fun(*unwrap_tuple(point)), value))

        return atoms

    def __getitem__(self, arg):
        try:
            expr, sigma = arg
            return self.evaluator(expr, self, sigma)
        except ValueError:
            # MRJ: This for expression that have the __getitem__ operator overloaded
            return self.evaluator(arg, self)
        except TypeError:
            return self.evaluator(arg, self)


def create(lang):
    return Model(lang)


class ExtensionalFunctionDefinition:
    def __init__(self):
        self.data = dict()

    def set(self, point, value):
        assert isinstance(point, tuple)
        assert isinstance(value, Constant)
        self.data[wrap_tuple(point)] = value

    def get(self, point):
        return self.data[wrap_tuple(point)]


# class IntensionalFunctionDefinition:
#     pass

def wrap_tuple(tup):
    """ Create a tuple of Term references from a tuple of terms """
    return tuple(symref(a) for a in tup)


def unwrap_tuple(tup):
    """ Create a tuple of Tarski terms from a tuple of Term references"""
    return tuple(ref.expr for ref in tup)
